%{

/*#define YYDEBUG 1*/
/*-------------------------------------------------------------------------
 *
 * gram.y
 *	  POSTGRES SQL YACC rules/actions
 *
 * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *	  $Header: /cvsroot/pgsql/src/backend/parser/gram.y,v 2.436.2.1 2003/11/24 16:54:15 tgl Exp $
 *
 * HISTORY
 *	  AUTHOR			DATE			MAJOR EVENT
 *	  Andrew Yu			Sept, 1994		POSTQUEL to SQL conversion
 *	  Andrew Yu			Oct, 1994		lispy code conversion
 *
 * NOTES
 *	  CAPITALS are used to represent terminal symbols.
 *	  non-capitals are used to represent non-terminals.
 *	  SQL92-specific syntax is separated from plain SQL/Postgres syntax
 *	  to help isolate the non-extensible portions of the parser.
 *
 *	  In general, nothing in this file should initiate database accesses
 *	  nor depend on changeable state (such as SET variables).  If you do
 *	  database accesses, your code will fail when we have aborted the
 *	  current transaction and are just parsing commands to find the next
 *	  ROLLBACK or COMMIT.  If you make use of SET variables, then you
 *	  will do the wrong thing in multi-query strings like this:
 *			SET SQL_inheritance TO off; SELECT * FROM foo;
 *	  because the entire string is parsed by gram.y before the SET gets
 *	  executed.  Anything that depends on the database or changeable state
 *	  should be handled inside parse_analyze() so that it happens at the
 *	  right time not the wrong time.  The handling of SQL_inheritance is
 *	  a good example.
 *
 * WARNINGS
 *	  If you use a list, make sure the datum is a node so that the printing
 *	  routines work.
 *
 *	  Sometimes we assign constants to makeStrings. Make sure we don't free
 *	  those.
 *
 *-------------------------------------------------------------------------
 */

#define IFSQL_SCHEMA_NAME "_ifs"
#define IFSQL_TEMPLATE_NAME "_ifs_template"
#define IFSQL_SET_OPERATION_SUB_SELECT "IFSQL_SET_OPERATION_SUB_SELECT"
#define IFSQL_UNION_COUNT "IFSQL_UNION_COUNT"

#define IFSQL_COLUMNS \
    "mship  double precision DEFAULT 1 NOT NULL CHECK ( mship BETWEEN 0 AND 1), " \
    "nmship double precision DEFAULT 0 NOT NULL CHECK (nmship BETWEEN 0 AND 1), " \
    "CHECK ((mship + nmship) BETWEEN 0 AND 1), "

#define DEFAULT_TRANSLATION_ACTION \
	if (yylen == 0) yyval.str = ""; \
	else { yyval.str = yyvsp[1-yylen].str; for (int i = 1; i < yylen; ++i) yyval.str += " " + yyvsp[i+1-yylen].str; }

#include <iostream>
#include <string>
#include <deque>
using namespace std;

#include "gramparse.hpp"

extern string translated;			/* final translation result is delivered here */

enum StmtType {stUnknown, stSelect};
enum StmtContext {scUnknown, scWhere, scHaving };
struct StmtVars
{
	StmtType stmtType;
	StmtContext stmtContext;
	deque <string> selectList;
	deque <string> selectAliasList;
	deque <string> fromList;
	string membership;
	string tableAlias;
	
	deque <string> setOperAliasList;
	bool setOperFlag;
	StmtVars* parent;
	
	void createFromClauseMembership ();
	void modifyMembership (const string& modifierName, const string& modifierParams);
	string createTargetList ();
	void createUnionTargetList (string& targetList, string& groupByList);
	void createIntersectTargetList (string& targetList, string& groupByList);
};
deque <StmtVars> stmtVarsStack;

static void pushStmtVars (StmtType stmtType);
static void popStmtVars ();
static StmtVars& getStmtVars ();

static string concatenateStrings (const deque <string>& stringList);
static string modifierFunctionCall (const string& membership, const string& modifierName, const string& modifierParams);
static bool isRelationCrisp (const string& relname);

/*
 * If you need access to certain yacc-generated variables and find that
 * they're static by default, uncomment the next line.  (this is not a
 * problem, yet.)
 */
/*#define __YYSCLASS*/

%}


%union
{
	string str;
}

%type <str>	stmt schema_stmt
		AlterDatabaseSetStmt AlterDomainStmt AlterGroupStmt
		AlterSeqStmt AlterTableStmt AlterUserStmt AlterUserSetStmt
		AnalyzeStmt ClosePortalStmt ClusterStmt CommentStmt
		ConstraintsSetStmt CopyStmt CreateAsStmt CreateCastStmt
		CreateDomainStmt CreateGroupStmt CreateOpClassStmt CreatePLangStmt
		CreateSchemaStmt CreateSeqStmt CreateStmt
		CreateAssertStmt CreateTrigStmt CreateUserStmt
		CreatedbStmt DBTypeSetStmt DBTypeShowStmt
		DeclareCursorStmt DefineStmt DeleteStmt
		DropGroupStmt DropOpClassStmt DropPLangStmt DropStmt
		DropAssertStmt DropTrigStmt DropRuleStmt DropCastStmt
		DropUserStmt DropdbStmt ExplainStmt FetchStmt
		GrantStmt IndexStmt InsertStmt ListenStmt LoadStmt
		LockStmt NotifyStmt ExplainableStmt PreparableStmt
		CreateFunctionStmt ReindexStmt RemoveAggrStmt
		RemoveFuncStmt RemoveModifStmt RemovePredStmt
		RemoveOperStmt RenameStmt RevokeStmt
		RuleActionStmt RuleActionStmtOrEmpty RuleStmt
		SelectStmt TransactionStmt TruncateStmt
		UnlistenStmt UpdateStmt VacuumStmt
		VariableResetStmt VariableSetStmt VariableShowStmt
		ViewStmt CheckPointStmt CreateConversionStmt
		DeallocateStmt PrepareStmt ExecuteStmt
		CreatePredicateStmt
		CreateModifierStmt

%type <str>	select_no_parens select_with_parens select_clause
				simple_select

%type <str>	alter_column_default opclass_item
%type <str>	add_drop

%type <str>	opt_drop_behavior

%type <str>	createdb_opt_list copy_opt_list
%type <str>	createdb_opt_item copy_opt_item

%type <str>	opt_lock lock_type cast_context
%type <str>	opt_force opt_or_replace transaction_access_mode
				opt_grant_grant_option opt_revoke_grant_option

%type <str>	like_including_defaults

%type <str>	user_list

%type <str>	OptGroupList
%type <str>	OptGroupElem

%type <str>	OptUserList
%type <str>	OptUserElem

%type <str>		OptSchemaName
%type <str>	OptSchemaEltList

%type <str> TriggerActionTime TriggerForSpec opt_trusted
%type <str>		opt_lancompiler

%type <str>		TriggerEvents
%type <str>	TriggerFuncArg

%type <str>		relation_name copy_file_name
				database_name access_method_clause access_method attr_name
				index_name name function_name file_name

%type <str>	func_name handler_name qual_Op qual_all_Op
				opt_class opt_validator

%type <str>	qualified_name OptConstrFromTable

%type <str>		all_Op MathOp opt_name SpecialRuleRelation

%type <str>		iso_level opt_encoding
%type <str>	grantee
%type <str>	grantee_list
%type <str>	privilege
%type <str>	privileges privilege_list
%type <str> privilege_target
%type <str>	function_with_argtypes
%type <str>	function_with_argtypes_list
%type <str> 	TriggerOneEvent

%type <str>	stmtblock stmtmulti
				OptTableElementList TableElementList OptInherit definition
				opt_distinct opt_definition func_args
				func_args_list func_as createfunc_opt_list
				oper_argtypes RuleActionList RuleActionMulti
				opt_column_list columnList opt_name_list
				sort_clause opt_sort_clause sortby_list index_params
				name_list from_clause from_list opt_array_bounds
				qualified_name_list any_name any_name_list
				any_operator expr_list dotted_name attrs
				target_list update_target_list insert_column_list
				insert_target_list def_list opt_indirection
				group_clause TriggerFuncArgs select_limit
				opt_select_limit opclass_item_list transaction_mode_list
				transaction_mode_list_or_empty
				TableFuncElementList
				prep_type_clause prep_type_list
				execute_param_clause
				
				modify_clause
				modifier_list
				modifier_args
				select_recognizer
				select_where_recognizer
				select_having_recognizer
				set_operation_recognizer
				select_target_el
				opt_expr_list
				
%type <str>	into_clause OptTempTableName

%type <str>	createfunc_opt_item
%type <str>	func_arg func_return func_type aggr_argtype

%type <str> opt_arg TriggerForType OptTemp OptWithOids
%type <str>	OnCommitOption

%type <str>	for_update_clause opt_for_update_clause update_list
%type <str>	opt_all

%type <str>	join_outer join_qual
%type <str>	join_type

%type <str>	extract_list overlay_list position_list
%type <str>	substr_list trim_list
%type <str>	opt_interval
%type <str>	overlay_placing substr_from substr_for

%type <str> opt_instead opt_analyze
%type <str> index_opt_unique opt_verbose opt_full
%type <str> opt_freeze opt_default opt_recheck
%type <str>	opt_binary opt_oids copy_delimiter

%type <str> copy_from opt_hold

%type <str>	fetch_count	opt_column event cursor_options
%type <str>	reindex_type drop_type comment_type

%type <str>	fetch_direction select_limit_value select_offset_value

%type <str>	OptSeqList
%type <str>	OptSeqElem

%type <str>	insert_rest

%type <str> set_rest

%type <str>	TableElement ConstraintElem TableFuncElement
%type <str>	columnDef
%type <str>	def_elem
%type <str>	def_arg columnElem where_clause insert_column_item
				a_expr b_expr c_expr r_expr AexprConst
				in_expr having_clause func_table array_expr
%type <str>	row row_descriptor type_list array_expr_list
%type <str>	case_expr case_arg when_clause case_default
%type <str>	when_clause_list
%type <str>	sub_type
%type <str>	OptCreateAs CreateAsList
%type <str>	CreateAsElement
%type <str>	NumericOnly FloatOnly IntegerOnly
%type <str>	columnref
%type <str>	alias_clause
%type <str>	sortby
%type <str>	index_elem
%type <str>	table_ref
%type <str>	joined_table
%type <str>	relation_expr
%type <str>	target_el insert_target_el update_target_el

%type <str>	Typename SimpleTypename ConstTypename
				GenericType Numeric opt_float
				Character ConstCharacter
				CharacterWithLength CharacterWithoutLength
				ConstDatetime ConstInterval
				Bit ConstBit BitWithLength BitWithoutLength
%type <str>		character
%type <str>		extract_arg
%type <str>		opt_charset
%type <str>	opt_numeric opt_decimal
%type <str> opt_varying opt_timezone

%type <str>	Iconst
%type <str>		Sconst comment_text
%type <str>		UserId opt_boolean ColId_or_Sconst
%type <str>	var_list var_list_or_default
%type <str>		ColId ColLabel type_name
%type <str>	var_value zone_value

%type <str> unreserved_keyword func_name_keyword
%type <str> col_name_keyword reserved_keyword

%type <str>	TableConstraint TableLikeClause
%type <str>	ColQualList
%type <str>	ColConstraint ColConstraintElem ConstraintAttr
%type <str>	key_actions key_delete key_match key_update key_action
%type <str>	ConstraintAttributeSpec ConstraintDeferrabilitySpec
				ConstraintTimeSpec

%type <str>	constraints_set_list
%type <str> constraints_set_mode


/*
 * If you make any token changes, update the keyword table in
 * parser/keywords.c and add new keywords to the appropriate one of
 * the reserved-or-not-so-reserved keyword lists, below.
 */

/* ordinary key words in alphabetical order */
%token <str> ABORT_P ABSOLUTE_P ACCESS ACTION ADD AFTER
	AGGREGATE ALL ALTER ANALYSE ANALYZE AND ANY ARRAY AS ASC
	ASSERTION ASSIGNMENT AT AUTHORIZATION

	BACKWARD BEFORE BEGIN_P BETWEEN BIGINT BINARY BIT
	BOOLEAN_P BOTH BY

	CACHE CALLED CASCADE CASE CAST CHAIN CHAR_P
	CHARACTER CHARACTERISTICS CHECK CHECKPOINT CLASS CLOSE
	CLUSTER COALESCE COLLATE COLUMN COMMENT COMMIT
	COMMITTED CONSTRAINT CONSTRAINTS CONVERSION_P CONVERT COPY CREATE CREATEDB
	CREATEUSER CROSS CURRENT_DATE CURRENT_TIME
	CURRENT_TIMESTAMP CURRENT_USER CURSOR CYCLE

	DATABASE DAY_P DEALLOCATE DEC DECIMAL_P DECLARE DEFAULT DEFAULTS
	DEFERRABLE DEFERRED DEFINER DELETE_P DELIMITER DELIMITERS
    DESC DISTINCT DO DOMAIN_P DOUBLE_P DROP

	EACH ELSE ENCODING ENCRYPTED END_P ESCAPE EXCEPT EXCLUDING
	EXCLUSIVE EXECUTE EXISTS EXPLAIN EXTERNAL EXTRACT

	FALSE_P FETCH FIRST_P FLOAT_P FOR FORCE FOREIGN FORWARD
	FREEZE FROM FULL FUNCTION

	GLOBAL GRANT GROUP_P

	HANDLER HAVING HOLD HOUR_P

	ILIKE IMMEDIATE IMMUTABLE IMPLICIT_P IN_P INCLUDING INCREMENT
	INDEX INHERITS INITIALLY INNER_P INOUT INPUT_P
	INSENSITIVE INSERT INSTEAD INT_P INTEGER INTERSECT
	INTERVAL INTO INVOKER IS ISNULL ISOLATION

	JOIN

	KEY

	LANCOMPILER LANGUAGE LAST_P LEADING LEFT LEVEL LIKE LIMIT
	LISTEN LOAD LOCAL LOCALTIME LOCALTIMESTAMP LOCATION
	LOCK_P

	MATCH MAXVALUE MINUTE_P MINVALUE MODE MODIFIER MODIFY MONTH_P MOVE MSHIP

	NAMES NATIONAL NATURAL NCHAR NEW NEXT NMSHIP NO NOCREATEDB
	NOCREATEUSER NONE NOT NOTHING NOTIFY NOTNULL NULL_P
	NULLIF NUMERIC

	OF OFF OFFSET OIDS OLD ON ONLY OPERATOR OPTION OR
	ORDER OUT_P OUTER_P OVERLAPS OVERLAY OWNER

	PARTIAL PASSWORD PATH_P PENDANT PLACING POSITION
	PRECISION PREDICATE PRESERVE PREPARE PRIMARY 
	PRIOR PRIVILEGES PROCEDURAL PROCEDURE

	READ REAL RECHECK REFERENCES REINDEX RELATIVE_P RENAME REPLACE
	RESET RESTART RESTRICT RETURNS REVOKE RIGHT ROLLBACK ROW ROWS
	RULE

	SCHEMA SCROLL SECOND_P SECURITY SELECT SEQUENCE
	SERIALIZABLE SESSION SESSION_USER SET SETOF SHARE
	SHOW SIMILAR SIMPLE SMALLINT SOME STABLE START STATEMENT
	STATISTICS STDIN STDOUT STORAGE STRICT_P SUBSTRING SYSID

	TABLE TEMP TEMPLATE TEMPORARY THEN TIME TIMESTAMP
	TO TOAST TRAILING TRANSACTION TREAT TRIGGER TRIM TRUE_P
	TRUNCATE TRUSTED TYPE_P

	UNENCRYPTED UNION UNIQUE UNKNOWN UNLISTEN UNTIL
	UPDATE USAGE USER USING

	VACUUM VALID VALIDATOR VALUES VARCHAR VARYING
	VERBOSE VERSION VIEW VOLATILE

	WHEN WHERE WITH WITHOUT WORK WRITE

	YEAR_P

	ZONE

/* The grammar thinks these are keywords, but they are not in the keywords.c
 * list and so can never be entered directly.  The filter in parser.c
 * creates these tokens when required.
 */
%token			UNIONJOIN

/* Special keywords, not in the query language - see the "lex" file */
%token <str>	IDENT FCONST SCONST BCONST XCONST Op
%token <str>	ICONST PARAM

/* precedence: lowest to highest */
%left		UNION EXCEPT
%left		INTERSECT
%left		OR
%left		AND
%right		NOT
%right		'='
%nonassoc	'<' '>'
%nonassoc	LIKE ILIKE SIMILAR
%nonassoc	ESCAPE
%nonassoc	OVERLAPS
%nonassoc	BETWEEN
%nonassoc	IN_P
%left		POSTFIXOP		/* dummy for postfix Op rules */
%left		Op OPERATOR		/* multi-character ops and user-defined operators */
%nonassoc	NOTNULL
%nonassoc	ISNULL
%nonassoc	IS NULL_P TRUE_P FALSE_P UNKNOWN /* sets precedence for IS NULL, etc */
%left		'+' '-'
%left		'*' '/' '%'
%left		'^'
/* Unary Operators */
%left		AT ZONE			/* sets precedence for AT TIME ZONE */
%right		UMINUS
%left		'[' ']'
%left		'(' ')'
%left		TYPECAST
%left		'.'
/*
 * These might seem to be low-precedence, but actually they are not part
 * of the arithmetic hierarchy at all in their use as JOIN operators.
 * We make them high-precedence to support their use as function names.
 * They wouldn't be given a precedence at all, were it not that we need
 * left-associativity among the JOIN rules themselves.
 */
%left		JOIN UNIONJOIN CROSS LEFT FULL RIGHT INNER_P NATURAL
%%

/*
 *	Handle comment-only lines, and ;; SELECT * FROM pg_class ;;;
 *	psql already handles such cases, but other interfaces don't.
 *	bjm 1999/10/05
 */
stmtblock:	stmtmulti { translated = $1; }
		;

/* the thrashing around here is to discard "empty" statements... */
stmtmulti:	stmtmulti ';' stmt
				{}
			| stmt
					{}
		;

stmt :
			AlterDatabaseSetStmt
			| AlterDomainStmt
			| AlterGroupStmt
			| AlterSeqStmt
			| AlterTableStmt
			| AlterUserSetStmt
			| AlterUserStmt
			| AnalyzeStmt
			| CheckPointStmt
			| ClosePortalStmt
			| ClusterStmt
			| CommentStmt
			| ConstraintsSetStmt
			| CopyStmt
			| CreateAsStmt
			| CreateAssertStmt
			| CreateCastStmt
			| CreateConversionStmt
			| CreateDomainStmt
			| CreateFunctionStmt
			| CreatePredicateStmt
			| CreateModifierStmt
			| CreateGroupStmt
			| CreateOpClassStmt
			| CreatePLangStmt
			| CreateSchemaStmt
			| CreateSeqStmt
			| CreateStmt
			| CreateTrigStmt
			| CreateUserStmt
			| CreatedbStmt
			| DBTypeSetStmt
			| DBTypeShowStmt
			| DeallocateStmt
			| DeclareCursorStmt
			| DefineStmt
			| DeleteStmt
			| DropAssertStmt
			| DropCastStmt
			| DropGroupStmt
			| DropOpClassStmt
			| DropPLangStmt
			| DropRuleStmt
			| DropStmt
			| DropTrigStmt
			| DropUserStmt
			| DropdbStmt
			| ExecuteStmt
			| ExplainStmt
			| FetchStmt
			| GrantStmt
			| IndexStmt
			| InsertStmt
			| ListenStmt
			| LoadStmt
			| LockStmt
			| NotifyStmt
			| PrepareStmt
			| ReindexStmt
			| RemoveAggrStmt
			| RemoveFuncStmt
			| RemoveModifStmt
			| RemovePredStmt
			| RemoveOperStmt
			| RenameStmt
			| RevokeStmt
			| RuleStmt
			| SelectStmt
			| TransactionStmt
			| TruncateStmt
			| UnlistenStmt
			| UpdateStmt
			| VacuumStmt
			| VariableResetStmt
			| VariableSetStmt
			| VariableShowStmt
			| ViewStmt
			| /*EMPTY*/
				{}
		;

/*****************************************************************************
 *
 * Create a new Postgres DBMS user
 *
 *
 *****************************************************************************/

CreateUserStmt:
			CREATE USER UserId opt_with OptUserList
				{}
		;


opt_with:	WITH									{}
			| /*EMPTY*/								{}
		;

/*****************************************************************************
 *
 * Alter a postgresql DBMS user
 *
 *
 *****************************************************************************/

AlterUserStmt:
			ALTER USER UserId opt_with OptUserList
				 {}
		;


AlterUserSetStmt:
			ALTER USER UserId SET set_rest
				{}
			| ALTER USER UserId VariableResetStmt
				{}
			;


/*****************************************************************************
 *
 * Drop a postgresql DBMS user
 *
 * XXX Ideally this would have CASCADE/RESTRICT options, but since a user
 * might own objects in multiple databases, there is presently no way to
 * implement either cascading or restricting.  Caveat DBA.
 *****************************************************************************/

DropUserStmt:
			DROP USER user_list
				{}
			;

/*
 * Options for CREATE USER and ALTER USER
 */
OptUserList:
			OptUserList OptUserElem					{}
			| /* EMPTY */							{}
		;

OptUserElem:
			PASSWORD Sconst
				{}
			| ENCRYPTED PASSWORD Sconst
				{}
			| UNENCRYPTED PASSWORD Sconst
				{}
			| SYSID Iconst
				{}
			| CREATEDB
				{}
			| NOCREATEDB
				{}
			| CREATEUSER
				{}
			| NOCREATEUSER
				{}
			| IN_P GROUP_P user_list
				{}
			| VALID UNTIL Sconst
				{}
		;

user_list:	user_list ',' UserId		{}
			| UserId					{}
		;



/*****************************************************************************
 *
 * Create a postgresql group
 *
 *
 *****************************************************************************/

CreateGroupStmt:
			CREATE GROUP_P UserId opt_with OptGroupList
				{}
		;

/*
 * Options for CREATE GROUP
 */
OptGroupList:
			OptGroupList OptGroupElem				{}
			| /* EMPTY */							{}
		;

OptGroupElem:
			USER user_list
				{}
			| SYSID Iconst
				{}
		;


/*****************************************************************************
 *
 * Alter a postgresql group
 *
 *
 *****************************************************************************/

AlterGroupStmt:
			ALTER GROUP_P UserId add_drop USER user_list
				{}
		;

add_drop:	ADD										{}
			| DROP									{}
		;


/*****************************************************************************
 *
 * Drop a postgresql group
 *
 * XXX see above notes about cascading DROP USER; groups have same problem.
 *****************************************************************************/

DropGroupStmt:
			DROP GROUP_P UserId
				{}
		;


/*****************************************************************************
 *
 * Manipulate a schema
 *
 *****************************************************************************/

CreateSchemaStmt:
			CREATE SCHEMA OptSchemaName AUTHORIZATION UserId OptSchemaEltList
				{}
			| CREATE SCHEMA ColId OptSchemaEltList
				{}
		;

OptSchemaName:
			ColId									{}
			| /* EMPTY */							{}
		;

OptSchemaEltList:
			OptSchemaEltList schema_stmt			{}
			| /* EMPTY */							{}
		;

/*
 *	schema_stmt are the ones that can show up inside a CREATE SCHEMA
 *	statement (in addition to by themselves).
 */
schema_stmt:
			CreateStmt
			| GrantStmt
			| ViewStmt
		;

/*******************************************
 DATABASE TYPE manipulation statements
 Possible values are:
   fuzzy
   probabilistic
 ******************************************/

DBTypeSetStmt:
			SET DATABASE TYPE_P '=' IDENT
			{
				$$ = string() + "select "IFSQL_SCHEMA_NAME".set_dbtype ('" + $5 + "')"
					+ " as \"DATABASE TYPE\"";
			}
		;

DBTypeShowStmt:
			SHOW DATABASE TYPE_P
			{
				$$ = string() + "select "IFSQL_SCHEMA_NAME".show_dbtype ()"
					+ " as \"DATABASE TYPE\"";
			}
		;

/*****************************************************************************
 *
 * Set PG internal variable
 *	  SET name TO 'var_value'
 * Include SQL92 syntax (thomas 1997-10-22):
 *	  SET TIME ZONE 'var_value'
 *
 *****************************************************************************/

VariableSetStmt:
			SET set_rest
				{}
			| SET LOCAL set_rest
				{}
			| SET SESSION set_rest
				{}
		;

set_rest:  ColId TO var_list_or_default
				{}
			| ColId '=' var_list_or_default
				{}
			| TIME ZONE zone_value
				{}
			| TRANSACTION transaction_mode_list
				{}
			| SESSION CHARACTERISTICS AS TRANSACTION transaction_mode_list
				{}
			| NAMES opt_encoding
				{}
			| SESSION AUTHORIZATION ColId_or_Sconst
				{}
			| SESSION AUTHORIZATION DEFAULT
				{}
		;

var_list_or_default:
			var_list								{}
			| DEFAULT								{}
		;

var_list:	var_value								{}
			| var_list ',' var_value				{}
		;

var_value:	opt_boolean
				{}
			| ColId_or_Sconst
				{}
			| NumericOnly
				{}
		;

iso_level:	READ COMMITTED							{}
			| SERIALIZABLE							{}
		;

opt_boolean:
			TRUE_P									{}
			| FALSE_P								{}
			| ON									{}
			| OFF									{}
		;

/* Timezone values can be:
 * - a string such as 'pst8pdt'
 * - an identifier such as "pst8pdt"
 * - an integer or floating point number
 * - a time interval per SQL99
 * ColId gives reduce/reduce errors against ConstInterval and LOCAL,
 * so use IDENT and reject anything which is a reserved word.
 */
zone_value:
			Sconst
				{}
			| IDENT
				{}
			| ConstInterval Sconst opt_interval
				{}
			| ConstInterval '(' Iconst ')' Sconst opt_interval
				{}
			| NumericOnly							{}
			| DEFAULT								{}
			| LOCAL									{}
		;

opt_encoding:
			Sconst									{}
			| DEFAULT								{}
			| /*EMPTY*/								{}
		;

ColId_or_Sconst:
			ColId									{}
			| SCONST								{}
		;


VariableShowStmt:
			SHOW ColId
				{}
			| SHOW TIME ZONE
				{}
			| SHOW TRANSACTION ISOLATION LEVEL
				{}
			| SHOW SESSION AUTHORIZATION
				{}
			| SHOW ALL
				{}
		;

VariableResetStmt:
			RESET ColId
				{}
			| RESET TIME ZONE
				{}
			| RESET TRANSACTION ISOLATION LEVEL
				{}
			| RESET SESSION AUTHORIZATION
				{}
			| RESET ALL
				{}
		;


ConstraintsSetStmt:
			SET CONSTRAINTS constraints_set_list constraints_set_mode
				{}
		;

constraints_set_list:
			ALL										{}
			| name_list								{}
		;

constraints_set_mode:
			DEFERRED								{}
			| IMMEDIATE								{}
		;


/*
 * Checkpoint statement
 */
CheckPointStmt:
			CHECKPOINT
				{}
		;


/*****************************************************************************
 *
 *	ALTER TABLE variations
 *
 *****************************************************************************/

AlterTableStmt:
			/* ALTER TABLE <relation> ADD [COLUMN] <coldef> */
			ALTER TABLE relation_expr ADD opt_column columnDef
				{}
			/* ALTER TABLE <relation> ALTER [COLUMN] <colname> {} */
			| ALTER TABLE relation_expr ALTER opt_column ColId alter_column_default
				{}
			/* ALTER TABLE <relation> ALTER [COLUMN] <colname> DROP NOT NULL */
			| ALTER TABLE relation_expr ALTER opt_column ColId DROP NOT NULL_P
				{}
			/* ALTER TABLE <relation> ALTER [COLUMN] <colname> SET NOT NULL */
			| ALTER TABLE relation_expr ALTER opt_column ColId SET NOT NULL_P
				{}
			/* ALTER TABLE <relation> ALTER [COLUMN] <colname> SET STATISTICS <IntegerOnly> */
			| ALTER TABLE relation_expr ALTER opt_column ColId SET STATISTICS IntegerOnly
				{}
			/* ALTER TABLE <relation> ALTER [COLUMN] <colname> SET STORAGE <storagemode> */
			| ALTER TABLE relation_expr ALTER opt_column ColId
			SET STORAGE ColId
				{}
			/* ALTER TABLE <relation> DROP [COLUMN] <colname> [RESTRICT|CASCADE] */
			| ALTER TABLE relation_expr DROP opt_column ColId opt_drop_behavior
				{}
			/* ALTER TABLE <relation> ADD CONSTRAINT ... */
			| ALTER TABLE relation_expr ADD TableConstraint
				{}
			/* ALTER TABLE <relation> DROP CONSTRAINT <name> [RESTRICT|CASCADE] */
			| ALTER TABLE relation_expr DROP CONSTRAINT name opt_drop_behavior
				{}
			/* ALTER TABLE <relation> SET WITHOUT OIDS  */
			| ALTER TABLE relation_expr SET WITHOUT OIDS
				{}
			/* ALTER TABLE <name> CREATE TOAST TABLE */
			| ALTER TABLE qualified_name CREATE TOAST TABLE
				{}
			/* ALTER TABLE <name> OWNER TO UserId */
			| ALTER TABLE qualified_name OWNER TO UserId
				{}
			/* ALTER TABLE <name> CLUSTER ON <indexname> */
			| ALTER TABLE qualified_name CLUSTER ON name
				{}
		;

alter_column_default:
			SET DEFAULT a_expr
				{}
			| DROP DEFAULT					{}
		;

opt_drop_behavior:
			CASCADE						{}
			| RESTRICT					{}
			| /* EMPTY */				{}
		;


/*****************************************************************************
 *
 *		QUERY :
 *				close <portalname>
 *
 *****************************************************************************/

ClosePortalStmt:
			CLOSE name
				{}
		;


/*****************************************************************************
 *
 *		QUERY :
 *				COPY <relname> ['(' columnList ')'] FROM/TO [WITH options]
 *
 *				BINARY, OIDS, and DELIMITERS kept in old locations
 *				for backward compatibility.  2002-06-18
 *
 *****************************************************************************/

CopyStmt:	COPY opt_binary qualified_name opt_column_list opt_oids
			copy_from copy_file_name copy_delimiter opt_with copy_opt_list
				{}
		;

copy_from:
			FROM									{}
			| TO									{}
		;

/*
 * copy_file_name NULL indicates stdio is used. Whether stdin or stdout is
 * used depends on the direction. (It really doesn't make sense to copy from
 * stdout. We silently correct the "typo".		 - AY 9/94
 */
copy_file_name:
			Sconst									{}
			| STDIN									{}
			| STDOUT								{}
		;



copy_opt_list:
			copy_opt_list copy_opt_item				{}
			| /* EMPTY */							{}
		;


copy_opt_item:
			BINARY
				{}
			| OIDS
				{}
			| DELIMITER opt_as Sconst
				{}
			| NULL_P opt_as Sconst
				{}
		;

/* The following exist for backward compatibility */

opt_binary:
			BINARY
				{}
			| /*EMPTY*/								{}
		;

opt_oids:
			WITH OIDS
				{}
			| /*EMPTY*/								{}
		;

copy_delimiter:
			/* USING DELIMITERS kept for backward compatibility. 2002-06-15 */
			opt_using DELIMITERS Sconst
				{}
			| /*EMPTY*/								{}
		;

opt_using:
			USING									{}
			| /*EMPTY*/								{}
		;


/*****************************************************************************
 *
 *		QUERY :
 *				CREATE TABLE relname
 *
 *****************************************************************************/

CreateStmt:	CREATE OptTemp TABLE qualified_name '(' OptTableElementList ')'
			OptInherit OptWithOids OnCommitOption
				{ $$ = $1 + " " + $2 + " " + $3 + " " + $4 
					+ " ( " + (!isRelationCrisp($4) ? IFSQL_COLUMNS : "") + $6 + " ) " 
					+ $8 + " " + $9 + " " + $10;
				}
		| CREATE OptTemp TABLE qualified_name OF qualified_name
			'(' OptTableElementList ')' OptWithOids OnCommitOption
				{ $$ = $1 + " " + $2 + " " + $3 + " " + $4 + " " + $5 + " " + $6
					+ " ( " + (!isRelationCrisp($4) ? IFSQL_COLUMNS : "") + $8 + " ) " 
					+ $10 + " " + $11;
				}
		;

/*
 * Redundancy here is needed to avoid shift/reduce conflicts,
 * since TEMP is not a reserved word.  See also OptTempTableName.
 *
 * NOTE: we accept both GLOBAL and LOCAL options; since we have no modules
 * the LOCAL keyword is really meaningless.
 */
OptTemp:	TEMPORARY						{}
			| TEMP							{}
			| LOCAL TEMPORARY				{}
			| LOCAL TEMP					{}
			| GLOBAL TEMPORARY				{}
			| GLOBAL TEMP					{}
			| /*EMPTY*/						{}
		;

OptTableElementList:
			TableElementList					{}
			| /*EMPTY*/							{}
		;

TableElementList:
			TableElement
				{}
			| TableElementList ',' TableElement
				{}
		;

TableElement:
			columnDef							{}
			| TableLikeClause					{}
			| TableConstraint					{}
		;

columnDef:	ColId Typename ColQualList
				{}
		;

ColQualList:
			ColQualList ColConstraint				{}
			| /*EMPTY*/								{}
		;

ColConstraint:
			CONSTRAINT name ColConstraintElem
				{}
			| ColConstraintElem						{}
			| ConstraintAttr						{}
		;

/* DEFAULT NULL is already the default for Postgres.
 * But define it here and carry it forward into the system
 * to make it explicit.
 * - thomas 1998-09-13
 *
 * WITH NULL and NULL are not SQL92-standard syntax elements,
 * so leave them out. Use DEFAULT NULL to explicitly indicate
 * that a column may have that value. WITH NULL leads to
 * shift/reduce conflicts with WITH TIME ZONE anyway.
 * - thomas 1999-01-08
 *
 * DEFAULT expression must be b_expr not a_expr to prevent shift/reduce
 * conflict on NOT (since NOT might start a subsequent NOT NULL constraint,
 * or be part of a_expr NOT LIKE or similar constructs).
 */
ColConstraintElem:
			NOT NULL_P
				{}
			| NULL_P
				{}
			| UNIQUE
				{}
			| PRIMARY KEY
				{}
			| CHECK '(' a_expr ')'
				{}
			| DEFAULT b_expr
				{}
			| REFERENCES qualified_name opt_column_list key_match key_actions
				{}
		;

/*
 * ConstraintAttr represents constraint attributes, which we parse as if
 * they were independent constraint clauses, in order to avoid shift/reduce
 * conflicts (since NOT might start either an independent NOT NULL clause
 * or an attribute).  analyze.c is responsible for attaching the attribute
 * information to the preceding "real" constraint node, and for complaining
 * if attribute clauses appear in the wrong place or wrong combinations.
 *
 * See also ConstraintAttributeSpec, which can be used in places where
 * there is no parsing conflict.
 */
ConstraintAttr:
			DEFERRABLE
				{}
			| NOT DEFERRABLE
				{}
			| INITIALLY DEFERRED
				{}
			| INITIALLY IMMEDIATE
				{}
		;


/*
 * SQL99 supports wholesale borrowing of a table definition via the LIKE clause.
 * This seems to be a poor man's inheritance capability, with the resulting
 * tables completely decoupled except for the original commonality in definitions.
 *
 * This is very similar to CREATE TABLE AS except for the INCLUDING DEFAULTS extension
 * which is a part of SQL 200N
 */
TableLikeClause:
			LIKE qualified_name like_including_defaults
				{}
		;

like_including_defaults:
				INCLUDING DEFAULTS		{}
				| EXCLUDING DEFAULTS		{}
				| /* EMPTY */				{}
		;


/* ConstraintElem specifies constraint syntax which is not embedded into
 *	a column definition. ColConstraintElem specifies the embedded form.
 * - thomas 1997-12-03
 */
TableConstraint:
			CONSTRAINT name ConstraintElem
				{}
			| ConstraintElem						{}
		;

ConstraintElem:
			CHECK '(' a_expr ')'
				{}
			| UNIQUE '(' columnList ')'
				{}
			| PRIMARY KEY '(' columnList ')'
				{}
			| FOREIGN KEY '(' columnList ')' REFERENCES qualified_name
				opt_column_list key_match key_actions ConstraintAttributeSpec
				{}
		;

opt_column_list:
			'(' columnList ')'						{}
			| /*EMPTY*/								{}
		;

columnList:
			columnElem								{}
			| columnList ',' columnElem				{}
		;

columnElem: ColId
				{}
		;

key_match:  MATCH FULL
			{}
		| MATCH PARTIAL
			{}
		| MATCH SIMPLE
			{}
		| /*EMPTY*/
			{}
		;

/*
 * We combine the update and delete actions into one value temporarily
 * for simplicity of parsing, and then break them down again in the
 * calling production.  update is in the left 8 bits, delete in the right.
 * Note that NOACTION is the default.
 */
key_actions:
			key_update
				{}
			| key_delete
				{}
			| key_update key_delete
				{}
			| key_delete key_update
				{}
			| /*EMPTY*/
				{}
		;

key_update: ON UPDATE key_action		{}
		;

key_delete: ON DELETE_P key_action		{}
		;

key_action:
			NO ACTION					{}
			| RESTRICT					{}
			| CASCADE					{}
			| SET NULL_P				{}
			| SET DEFAULT				{}
		;

OptInherit: INHERITS '(' qualified_name_list ')'
				{}
			| /*EMPTY*/
				{}
		;

OptWithOids:
			WITH OIDS								{}
			| WITHOUT OIDS							{}
			| /*EMPTY*/								{}
		;

OnCommitOption:  ON COMMIT DROP				{}
			| ON COMMIT DELETE_P ROWS		{}
			| ON COMMIT PRESERVE ROWS		{}
			| /*EMPTY*/						{}
		;


/*
 * Note: CREATE TABLE ... AS SELECT ... is just another spelling for
 * SELECT ... INTO.
 */

CreateAsStmt:
			CREATE OptTemp TABLE qualified_name OptCreateAs AS SelectStmt
				{}
		;

OptCreateAs:
			'(' CreateAsList ')'					{}
			| /*EMPTY*/								{}
		;

CreateAsList:
			CreateAsElement							{}
			| CreateAsList ',' CreateAsElement		{}
		;

CreateAsElement:
			ColId
				{}
		;


/*****************************************************************************
 *
 *		QUERY :
 *				CREATE SEQUENCE seqname
 *				ALTER SEQUENCE seqname
 *
 *****************************************************************************/

CreateSeqStmt:
			CREATE OptTemp SEQUENCE qualified_name OptSeqList
				{}
		;

AlterSeqStmt:
			ALTER SEQUENCE qualified_name OptSeqList
				{}
		;

OptSeqList: OptSeqList OptSeqElem					{}
			| /*EMPTY*/								{}
		;

OptSeqElem: CACHE NumericOnly
				{}
			| CYCLE
				{}
			| NO CYCLE
				{}
			| INCREMENT opt_by NumericOnly
				{}
			| MAXVALUE NumericOnly
				{}
			| MINVALUE NumericOnly
				{}
			| NO MAXVALUE
				{}
			| NO MINVALUE
				{}
			| START opt_with NumericOnly
				{}
			| RESTART opt_with NumericOnly
				{}
		;

opt_by:		BY				{}
			| /* empty */	{}
	  ;

NumericOnly:
			FloatOnly								{}
			| IntegerOnly							{}
		;

FloatOnly:	FCONST									{}
			| '-' FCONST
				{}
		;

IntegerOnly:
			Iconst
				{}
			| '-' Iconst
				{}
		;

/*****************************************************************************
 *
 *		QUERIES :
 *				CREATE PROCEDURAL LANGUAGE ...
 *				DROP PROCEDURAL LANGUAGE ...
 *
 *****************************************************************************/

CreatePLangStmt:
			CREATE opt_trusted opt_procedural LANGUAGE ColId_or_Sconst
			HANDLER handler_name opt_validator opt_lancompiler
			{}
		;

opt_trusted:
			TRUSTED									{}
			| /*EMPTY*/								{}
		;

/* This ought to be just func_name, but that causes reduce/reduce conflicts
 * (CREATE LANGUAGE is the only place where func_name isn't followed by '(').
 * Work around by using name and dotted_name separately.
 */
handler_name:
			name
							{}
			| dotted_name							{}
		;

opt_lancompiler:
			LANCOMPILER Sconst						{}
			| /*EMPTY*/								{}
		;

opt_validator:
			VALIDATOR handler_name {}
			| /*EMPTY*/ {}
		;

DropPLangStmt:
			DROP opt_procedural LANGUAGE ColId_or_Sconst opt_drop_behavior
				{}
		;

opt_procedural:
			PROCEDURAL								{}
			| /*EMPTY*/								{}
		;

/*****************************************************************************
 *
 *		QUERIES :
 *				CREATE TRIGGER ...
 *				DROP TRIGGER ...
 *
 *****************************************************************************/

CreateTrigStmt:
			CREATE TRIGGER name TriggerActionTime TriggerEvents ON
			qualified_name TriggerForSpec EXECUTE PROCEDURE
			func_name '(' TriggerFuncArgs ')'
				{}
			| CREATE CONSTRAINT TRIGGER name AFTER TriggerEvents ON
			qualified_name OptConstrFromTable
			ConstraintAttributeSpec
			FOR EACH ROW EXECUTE PROCEDURE
			func_name '(' TriggerFuncArgs ')'
				{}
		;

TriggerActionTime:
			BEFORE									{}
			| AFTER									{}
		;

TriggerEvents:
			TriggerOneEvent
				{}
			| TriggerOneEvent OR TriggerOneEvent
				{}
			| TriggerOneEvent OR TriggerOneEvent OR TriggerOneEvent
				{}
		;

TriggerOneEvent:
			INSERT									{}
			| DELETE_P								{}
			| UPDATE								{}
		;

TriggerForSpec:
			FOR TriggerForOpt TriggerForType
				{}
			| /* EMPTY */
				{}
		;

TriggerForOpt:
			EACH									{}
			| /*EMPTY*/								{}
		;

TriggerForType:
			ROW										{}
			| STATEMENT								{}
		;

TriggerFuncArgs:
			TriggerFuncArg							{}
			| TriggerFuncArgs ',' TriggerFuncArg	{}
			| /*EMPTY*/								{}
		;

TriggerFuncArg:
			ICONST
				{}
			| FCONST								{}
			| Sconst								{}
			| BCONST								{}
			| XCONST								{}
			| ColId									{}
		;

OptConstrFromTable:
			FROM qualified_name						{}
			| /*EMPTY*/								{}
		;

ConstraintAttributeSpec:
			ConstraintDeferrabilitySpec
				{}
			| ConstraintDeferrabilitySpec ConstraintTimeSpec
				{}
			| ConstraintTimeSpec
				{}
			| ConstraintTimeSpec ConstraintDeferrabilitySpec
				{}
			| /*EMPTY*/
				{}
		;

ConstraintDeferrabilitySpec:
			NOT DEFERRABLE							{}
			| DEFERRABLE							{}
		;

ConstraintTimeSpec:
			INITIALLY IMMEDIATE						{}
			| INITIALLY DEFERRED					{}
		;


DropTrigStmt:
			DROP TRIGGER name ON qualified_name opt_drop_behavior
				{}
		;


/*****************************************************************************
 *
 *		QUERIES :
 *				CREATE ASSERTION ...
 *				DROP ASSERTION ...
 *
 *****************************************************************************/

CreateAssertStmt:
			CREATE ASSERTION name CHECK '(' a_expr ')'
			ConstraintAttributeSpec
				{}
		;

DropAssertStmt:
			DROP ASSERTION name opt_drop_behavior
				{}
		;


/*****************************************************************************
 *
 *		QUERY :
 *				define (aggregate,operator,type)
 *
 *****************************************************************************/

DefineStmt:
			CREATE AGGREGATE func_name definition
				{}
			| CREATE OPERATOR any_operator definition
				{}
			| CREATE TYPE_P any_name definition
				{}
			| CREATE TYPE_P any_name AS '(' TableFuncElementList ')'
				{}
		;

definition: '(' def_list ')'						{}
		;

def_list:  	def_elem								{}
			| def_list ',' def_elem					{}
		;

def_elem:  ColLabel '=' def_arg
				{}
			| ColLabel
				{}
		;

/* Note: any simple identifier will be returned as a type name! */
def_arg:	func_return						{}
			| qual_all_Op					{}
			| NumericOnly					{}
			| Sconst						{}
		;


/*****************************************************************************
 *
 *		QUERIES :
 *				CREATE OPERATOR CLASS ...
 *				DROP OPERATOR CLASS ...
 *
 *****************************************************************************/

CreateOpClassStmt:
			CREATE OPERATOR CLASS any_name opt_default FOR TYPE_P Typename
			USING access_method AS opclass_item_list
				{}
		;

opclass_item_list:
			opclass_item							{}
			| opclass_item_list ',' opclass_item	{}
		;

opclass_item:
			OPERATOR Iconst any_operator opt_recheck
				{}
			| OPERATOR Iconst any_operator '(' oper_argtypes ')' opt_recheck
				{}
			| FUNCTION Iconst func_name func_args
				{}
			| STORAGE Typename
				{}
		;

opt_default:	DEFAULT	{}
			| /*EMPTY*/	{}
		;

opt_recheck:	RECHECK	{}
			| /*EMPTY*/	{}
		;


DropOpClassStmt:
			DROP OPERATOR CLASS any_name USING access_method opt_drop_behavior
				{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *
 *		DROP itemtype itemname [, itemname ...] [ RESTRICT | CASCADE ]
 *
 *****************************************************************************/

DropStmt:	DROP drop_type any_name_list opt_drop_behavior
				{}
		;

drop_type:	TABLE									{}
			| SEQUENCE								{}
			| VIEW									{}
			| INDEX									{}
			| TYPE_P								{}
			| DOMAIN_P								{}
			| CONVERSION_P							{}
			| SCHEMA								{}
		;

any_name_list:
			any_name								{}
			| any_name_list ',' any_name			{}
		;

any_name:	ColId						{}
			| dotted_name				{}
		;

/*****************************************************************************
 *
 *		QUERY:
 *				truncate table relname
 *
 *****************************************************************************/

TruncateStmt:
			TRUNCATE opt_table qualified_name
				{}
		;

/*****************************************************************************
 *
 *	The COMMENT ON statement can take different forms based upon the type of
 *	the object associated with the comment. The form of the statement is:
 *
 *	COMMENT ON [ [ DATABASE | DOMAIN | INDEX | SEQUENCE | TABLE | TYPE | VIEW ]
 *				 <objname> | AGGREGATE <aggname> (<aggtype>) | FUNCTION
 *		 <funcname> (arg1, arg2, ...) | OPERATOR <op>
 *		 (leftoperand_typ rightoperand_typ) | TRIGGER <triggername> ON
 *		 <relname> | RULE <rulename> ON <relname> ] IS 'text'
 *
 *****************************************************************************/

CommentStmt:
			COMMENT ON comment_type any_name IS comment_text
				{}
			| COMMENT ON AGGREGATE func_name '(' aggr_argtype ')'
			IS comment_text
				{}
			| COMMENT ON FUNCTION func_name func_args IS comment_text
				{}
			| COMMENT ON OPERATOR any_operator '(' oper_argtypes ')'
			IS comment_text
				{}
			| COMMENT ON CONSTRAINT name ON any_name IS comment_text
				{}
			| COMMENT ON RULE name ON any_name IS comment_text
				{}
			| COMMENT ON RULE name IS comment_text
				{}
			| COMMENT ON TRIGGER name ON any_name IS comment_text
				{}
		;

comment_type:
			COLUMN								{}
			| DATABASE							{}
			| SCHEMA							{}
			| INDEX								{}
			| SEQUENCE							{}
			| TABLE								{}
			| DOMAIN_P							{}
			| TYPE_P							{}
			| VIEW								{}
		;

comment_text:
			Sconst								{}
			| NULL_P							{}
		;

/*****************************************************************************
 *
 *		QUERY:
 *			fetch/move
 *
 *****************************************************************************/

FetchStmt:	FETCH fetch_direction from_in name
				{}
			| FETCH name
				{}
			| MOVE fetch_direction from_in name
				{}
			| MOVE name
				{}
		;

fetch_direction:
			/*EMPTY*/
				{}
			| NEXT
				{}
			| PRIOR
				{}
			| FIRST_P
				{}
			| LAST_P
				{}
			| ABSOLUTE_P fetch_count
				{}
			| RELATIVE_P fetch_count
				{}
			| fetch_count
				{}
			| ALL
				{}
			| FORWARD
				{}
			| FORWARD fetch_count
				{}
			| FORWARD ALL
				{}
			| BACKWARD
				{}
			| BACKWARD fetch_count
				{}
			| BACKWARD ALL
				{}
		;

fetch_count:
			Iconst									{}
			| '-' Iconst							{}
		;

from_in:	FROM									{}
			| IN_P									{}
		;


/*****************************************************************************
 *
 * GRANT and REVOKE statements
 *
 *****************************************************************************/

GrantStmt:	GRANT privileges ON privilege_target TO grantee_list
			opt_grant_grant_option
				{}
		;

RevokeStmt: REVOKE opt_revoke_grant_option privileges ON privilege_target
			FROM grantee_list opt_drop_behavior
				{}
		;


/* either ALL [PRIVILEGES] or a list of individual privileges */
privileges: privilege_list				{}
			| ALL						{}
			| ALL PRIVILEGES			{}
		;

privilege_list:
			privilege								{}
			| privilege_list ',' privilege			{}
		;

/* Not all of these privilege types apply to all objects, but that
 * gets sorted out later.
 */
privilege:	SELECT									{}
			| INSERT								{}
			| UPDATE								{}
			| DELETE_P								{}
			| RULE									{}
			| REFERENCES							{}
			| TRIGGER								{}
			| EXECUTE								{}
			| USAGE									{}
			| CREATE								{}
			| TEMPORARY								{}
			| TEMP									{}
		;


/* Don't bother trying to fold the first two rules into one using
   opt_table.  You're going to get conflicts. */
privilege_target:
			qualified_name_list
				{}
			| TABLE qualified_name_list
				{}
			| FUNCTION function_with_argtypes_list
				{}
			| DATABASE name_list
				{}
			| LANGUAGE name_list
				{}
			| SCHEMA name_list
				{}
		;


grantee_list:
			grantee									{}
			| grantee_list ',' grantee				{}
		;

grantee:	ColId
				{}
			| GROUP_P ColId
				{}
		;


opt_grant_grant_option:
			WITH GRANT OPTION {}
			| /*EMPTY*/ {}
		;

opt_revoke_grant_option:
			GRANT OPTION FOR {}
			| /*EMPTY*/ {}
		;


function_with_argtypes_list:
			function_with_argtypes					{}
			| function_with_argtypes_list ',' function_with_argtypes
													{}
		;

function_with_argtypes:
			func_name func_args
				{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				create index <indexname> on <relname>
 *				  [ using <access> ] "(" ( <col> [ using <opclass> ] )+ ")"
 *				  [ where <predicate> ]
 *
 *****************************************************************************/

IndexStmt:	CREATE index_opt_unique INDEX index_name ON qualified_name
			access_method_clause '(' index_params ')' where_clause
				{}
		;

index_opt_unique:
			UNIQUE									{}
			| /*EMPTY*/								{}
		;

access_method_clause:
			USING access_method						{}
			| /*EMPTY*/								{}
		;

index_params:	index_elem							{}
			| index_params ',' index_elem			{}
		;

/*
 * Index attributes can be either simple column references, or arbitrary
 * expressions in parens.  For backwards-compatibility reasons, we allow
 * an expression that's just a function call to be written without parens.
 */
index_elem:	attr_name opt_class
				{}
			| func_name '(' expr_list ')' opt_class
				{}
			| '(' a_expr ')' opt_class
				{}
		;

opt_class:	any_name								{}
			| USING any_name						{}
			| /*EMPTY*/								{}
		;

/*****************************************************************************
 *
 *		QUERY:
 *				create [or replace] function <fname>
 *						[(<type-1> {})]
 *						returns <type-r>
 *						as <filename or code in language as appropriate>
 *						language <lang> [with parameters]
 *
 *****************************************************************************/

CreatePredicateStmt:
			CREATE opt_or_replace PREDICATE func_name func_args
			createfunc_opt_list opt_definition
			{ $$ = $1 + " " + $2 + " function " + $4 + " " + $5 + 
				" returns " + IFSQL_SCHEMA_NAME".bool " + $6 + " " + $7; }
		;

CreateModifierStmt:
			CREATE opt_or_replace MODIFIER func_name modifier_args
			createfunc_opt_list opt_definition
			{ $$ = $1 + " " + $2 + " function " + $4 + " " + $5 + 
				" returns " + IFSQL_SCHEMA_NAME".bool " + $6 + " " + $7; }
		;

CreateFunctionStmt:
			CREATE opt_or_replace FUNCTION func_name func_args
			RETURNS func_return createfunc_opt_list opt_definition
				{}
		;

opt_or_replace:
			OR REPLACE								{}
			| /*EMPTY*/								{}
		;

modifier_args:	'(' func_args_list ')'					
			{ $$ = "("IFSQL_SCHEMA_NAME".bool, " + $2 + ")"; }
		| '(' ')'								
			{ $$ = "("IFSQL_SCHEMA_NAME".bool)"; }
		;

func_args:	'(' func_args_list ')'					{}
			| '(' ')'								{}
		;

func_args_list:
			func_arg								{}
			| func_args_list ',' func_arg			{}
		;

func_arg:	opt_arg func_type
				{}
			| func_type								{}
		;

opt_arg:	IN_P									{}
			| OUT_P
				{}
			| INOUT
				{}
		;

func_return:
			func_type
				{}
		;

/*
 * We would like to make the second production here be ColId attrs etc,
 * but that causes reduce/reduce conflicts.  type_name is next best choice.
 */
func_type:	Typename								{}
			| type_name attrs '%' TYPE_P
				{}
		;


createfunc_opt_list:
			/* Must be at least one to prevent conflict */
			createfunc_opt_item                     {}
			| createfunc_opt_list createfunc_opt_item {}
	;

createfunc_opt_item:
			AS func_as
				{}
			| LANGUAGE ColId_or_Sconst
				{}
			| IMMUTABLE
				{}
			| STABLE
				{}
			| VOLATILE
				{}
			| CALLED ON NULL_P INPUT_P
				{}
			| RETURNS NULL_P ON NULL_P INPUT_P
				{}
			| STRICT_P
				{}
			| EXTERNAL SECURITY DEFINER
				{}
			| EXTERNAL SECURITY INVOKER
				{}
			| SECURITY DEFINER
				{}
			| SECURITY INVOKER
				{}
		;

func_as:	Sconst						{}
			| Sconst ',' Sconst
				{}
		;

opt_definition:
			WITH definition							{}
			| /*EMPTY*/								{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *
 *		DROP FUNCTION funcname (arg1, arg2, ...) [ RESTRICT | CASCADE ]
 *		DROP AGGREGATE aggname (aggtype) [ RESTRICT | CASCADE ]
 *		DROP OPERATOR opname (leftoperand_typ, rightoperand_typ) [ RESTRICT | CASCADE ]
 *
 *****************************************************************************/

RemoveFuncStmt:
			DROP FUNCTION func_name func_args opt_drop_behavior
				{}
		;

RemovePredStmt:
			DROP PREDICATE func_name func_args opt_drop_behavior
			{ $$ = "DROP FUNCTION " + $3 + " " + $4 + " " + $5; }
		;

RemoveModifStmt:
			DROP MODIFIER func_name modifier_args opt_drop_behavior
			{ $$ = "DROP FUNCTION " + $3 + " " + $4 + " " + $5; }
		;

RemoveAggrStmt:
			DROP AGGREGATE func_name '(' aggr_argtype ')' opt_drop_behavior
				{}
		;

aggr_argtype:
			Typename								{}
			| '*'									{}
		;

RemoveOperStmt:
			DROP OPERATOR any_operator '(' oper_argtypes ')' opt_drop_behavior
				{}
		;

oper_argtypes:
			Typename
				{}
			| Typename ',' Typename
					{}
			| NONE ',' Typename /* left unary */
					{}
			| Typename ',' NONE /* right unary */
					{}
		;

any_operator:
			all_Op
					{}
			| ColId '.' any_operator
					{}
		;


/*****************************************************************************
 *
 *		CREATE CAST / DROP CAST
 *
 *****************************************************************************/

CreateCastStmt: CREATE CAST '(' Typename AS Typename ')'
					WITH FUNCTION function_with_argtypes cast_context
				{}
			| CREATE CAST '(' Typename AS Typename ')'
					WITHOUT FUNCTION cast_context
				{}
		;

cast_context:  AS IMPLICIT_P					{}
		| AS ASSIGNMENT							{}
		| /*EMPTY*/								{}
		;


DropCastStmt: DROP CAST '(' Typename AS Typename ')' opt_drop_behavior
				{}
		;



/*****************************************************************************
 *
 *		QUERY:
 *
 *		REINDEX type <typename> [FORCE] [ALL]
 *
 *****************************************************************************/

ReindexStmt:
			REINDEX reindex_type qualified_name opt_force
				{}
			| REINDEX DATABASE name opt_force
				{}
		;

reindex_type:
			INDEX									{}
			| TABLE									{}
		;

opt_force:	FORCE									{}
			| /* EMPTY */							{}
		;


/*****************************************************************************
 *
 * ALTER THING name RENAME TO newname
 *
 *****************************************************************************/

RenameStmt: ALTER AGGREGATE func_name '(' aggr_argtype ')' RENAME TO name
				{}
			| ALTER CONVERSION_P any_name RENAME TO name
				{}
			| ALTER DATABASE database_name RENAME TO database_name
				{}
			| ALTER FUNCTION func_name func_args RENAME TO name
				{}
			| ALTER GROUP_P UserId RENAME TO UserId
				{}
			| ALTER LANGUAGE name RENAME TO name
				{}
			| ALTER OPERATOR CLASS any_name USING access_method RENAME TO name
				{}
			| ALTER SCHEMA name RENAME TO name
				{}
			| ALTER TABLE relation_expr RENAME opt_column opt_name TO name
				{}
			| ALTER TRIGGER name ON relation_expr RENAME TO name
				{}
			| ALTER USER UserId RENAME TO UserId
				{}
		;

opt_name:	name									{}
			| /*EMPTY*/								{}
		;

opt_column: COLUMN									{}
			| /*EMPTY*/								{}
		;


/*****************************************************************************
 *
 *		QUERY:	Define Rewrite Rule
 *
 *****************************************************************************/

RuleStmt:	CREATE opt_or_replace RULE name AS
			{}
			ON event TO qualified_name where_clause
			DO opt_instead RuleActionList
				{}
		;

RuleActionList:
			NOTHING									{}
			| RuleActionStmt						{}
			| '(' RuleActionMulti ')'				{}
		;

/* the thrashing around here is to discard "empty" statements... */
RuleActionMulti:
			RuleActionMulti ';' RuleActionStmtOrEmpty
				{}
			| RuleActionStmtOrEmpty
				{}
		;

RuleActionStmt:
			SelectStmt
			| InsertStmt
			| UpdateStmt
			| DeleteStmt
			| NotifyStmt
		;

RuleActionStmtOrEmpty:
			RuleActionStmt							{}
			|	/*EMPTY*/							{}
		;

/* change me to select, update, etc. some day */
event:		SELECT									{}
			| UPDATE								{}
			| DELETE_P								{}
			| INSERT								{}
		 ;

opt_instead:
			INSTEAD									{}
			| /*EMPTY*/								{}
		;


DropRuleStmt:
			DROP RULE name ON qualified_name opt_drop_behavior
				{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				NOTIFY <qualified_name> can appear both in rule bodies and
 *				as a query-level command
 *
 *****************************************************************************/

NotifyStmt: NOTIFY qualified_name
				{}
		;

ListenStmt: LISTEN qualified_name
				{}
		;

UnlistenStmt:
			UNLISTEN qualified_name
				{}
			| UNLISTEN '*'
				{}
		;


/*****************************************************************************
 *
 *		Transactions:
 *
 *		BEGIN / COMMIT / ROLLBACK
 *		(also older versions END / ABORT)
 *
 *****************************************************************************/

TransactionStmt:
			ABORT_P opt_transaction
				{}
			| BEGIN_P opt_transaction
				{}
			| START TRANSACTION transaction_mode_list_or_empty
				{}
			| COMMIT opt_transaction
				{}
			| END_P opt_transaction
				{}
			| ROLLBACK opt_transaction
				{}
		;

opt_transaction:	WORK							{}
			| TRANSACTION							{}
			| /*EMPTY*/								{}
		;

transaction_mode_list:
			ISOLATION LEVEL iso_level
					{}
			| transaction_access_mode
					{}
			| ISOLATION LEVEL iso_level transaction_access_mode
					{}
			| transaction_access_mode ISOLATION LEVEL iso_level
					{}
		;

transaction_mode_list_or_empty:
			transaction_mode_list
			| /* EMPTY */
					{}
		;

transaction_access_mode:
			READ ONLY {}
			| READ WRITE {}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				create view <viewname> '('target-list ')' AS <query>
 *
 *****************************************************************************/

ViewStmt:	CREATE opt_or_replace VIEW qualified_name opt_column_list
				AS SelectStmt
				{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				load "filename"
 *
 *****************************************************************************/

LoadStmt:	LOAD file_name
				{}
		;


/*****************************************************************************
 *
 *		CREATE DATABASE
 *
 *****************************************************************************/

CreatedbStmt:
			CREATE DATABASE database_name opt_with createdb_opt_list
			{ $$ += " TEMPLATE = "IFSQL_TEMPLATE_NAME; }
		;

createdb_opt_list:
			createdb_opt_list createdb_opt_item		{}
			| /* EMPTY */							{}
		;

createdb_opt_item:
			LOCATION opt_equal Sconst
				{}
			| LOCATION opt_equal DEFAULT
				{}
			| ENCODING opt_equal Sconst
				{}
			| ENCODING opt_equal Iconst
				{}
			| ENCODING opt_equal DEFAULT
				{}
			| OWNER opt_equal name
				{}
			| OWNER opt_equal DEFAULT
				{}
		;

/*
 *	Though the equals sign doesn't match other WITH options, pg_dump uses
 *	equals for backward compability, and it doesn't seem worth removing it.
 *	2002-02-25
 */
opt_equal:	'='										{}
			| /*EMPTY*/								{}
		;


/*****************************************************************************
 *
 *		ALTER DATABASE
 *
 *****************************************************************************/

AlterDatabaseSetStmt:
			ALTER DATABASE database_name SET set_rest
				{}
			| ALTER DATABASE database_name VariableResetStmt
				{}
		;


/*****************************************************************************
 *
 *		DROP DATABASE
 *
 * This is implicitly CASCADE, no need for drop behavior
 *****************************************************************************/

DropdbStmt: DROP DATABASE database_name
				{}
		;


/*****************************************************************************
 *
 * Manipulate a domain
 *
 *****************************************************************************/

CreateDomainStmt:
			CREATE DOMAIN_P any_name opt_as Typename ColQualList
				{}
		;

AlterDomainStmt:
			/* ALTER DOMAIN <domain> {} */
			ALTER DOMAIN_P any_name alter_column_default
				{}
			/* ALTER DOMAIN <domain> DROP NOT NULL */
			| ALTER DOMAIN_P any_name DROP NOT NULL_P
				{}
			/* ALTER DOMAIN <domain> SET NOT NULL */
			| ALTER DOMAIN_P any_name SET NOT NULL_P
				{}
			/* ALTER DOMAIN <domain> ADD CONSTRAINT ... */
			| ALTER DOMAIN_P any_name ADD TableConstraint
				{}
			/* ALTER DOMAIN <domain> DROP CONSTRAINT <name> [RESTRICT|CASCADE] */
			| ALTER DOMAIN_P any_name DROP CONSTRAINT name opt_drop_behavior
				{}
			/* ALTER DOMAIN <domain> OWNER TO UserId */
			| ALTER DOMAIN_P any_name OWNER TO UserId
				{}
			;

opt_as:		AS										{}
			| /* EMPTY */							{}
		;


/*****************************************************************************
 *
 * Manipulate a conversion
 *
 *		CREATE [DEFAULT] CONVERSION <conversion_name>
 *		FOR <encoding_name> TO <encoding_name> FROM <func_name>
 *
 *****************************************************************************/

CreateConversionStmt:
			CREATE opt_default CONVERSION_P any_name FOR Sconst
			TO Sconst FROM any_name
			{}
		;

/*****************************************************************************
 *
 *		QUERY:
 *				cluster <index_name> on <qualified_name>
 *				cluster <qualified_name>
 *				cluster
 *
 *****************************************************************************/

ClusterStmt:
			CLUSTER index_name ON qualified_name
				{}
			| CLUSTER qualified_name
				{}
			| CLUSTER
			    {}
		;

/*****************************************************************************
 *
 *		QUERY:
 *				vacuum
 *				analyze
 *
 *****************************************************************************/

VacuumStmt: VACUUM opt_full opt_freeze opt_verbose
				{}
			| VACUUM opt_full opt_freeze opt_verbose qualified_name
				{}
			| VACUUM opt_full opt_freeze opt_verbose AnalyzeStmt
				{}
		;

AnalyzeStmt:
			analyze_keyword opt_verbose
				{}
			| analyze_keyword opt_verbose qualified_name opt_name_list
				{}
		;

analyze_keyword:
			ANALYZE									{}
			| ANALYSE /* British */					{}
		;

opt_verbose:
			VERBOSE									{}
			| /*EMPTY*/								{}
		;

opt_full:	FULL									{}
			| /*EMPTY*/								{}
		;

opt_freeze: FREEZE									{}
			| /*EMPTY*/								{}
		;

opt_name_list:
			'(' name_list ')'						{}
			| /*EMPTY*/								{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				EXPLAIN [ANALYZE] [VERBOSE] query
 *
 *****************************************************************************/

ExplainStmt: EXPLAIN opt_analyze opt_verbose ExplainableStmt
				{}
		;

ExplainableStmt:
			SelectStmt
			| InsertStmt
			| UpdateStmt
			| DeleteStmt
			| DeclareCursorStmt
			| ExecuteStmt					/* by default all are $$=$1 */
		;

opt_analyze:
			analyze_keyword			{}
			| /* EMPTY */			{}
		;

/*****************************************************************************
 *
 *		QUERY:
 *				PREPARE <plan_name> [(args, ...)] AS <query>
 *
 *****************************************************************************/

PrepareStmt: PREPARE name prep_type_clause AS PreparableStmt
				{}
		;

prep_type_clause: '(' prep_type_list ')'	{}
				| /* EMPTY */				{}
		;

prep_type_list: Typename			{}
			  | prep_type_list ',' Typename
									{}
		;

PreparableStmt:
			SelectStmt
			| InsertStmt
			| UpdateStmt
			| DeleteStmt					/* by default all are $$=$1 */
		;

/*****************************************************************************
 *
 * EXECUTE <plan_name> [(params, ...)]
 * CREATE TABLE <name> AS EXECUTE <plan_name> [(params, ...)]
 *
 *****************************************************************************/

ExecuteStmt: EXECUTE name execute_param_clause
				{}
			| CREATE OptTemp TABLE qualified_name OptCreateAs AS EXECUTE name execute_param_clause
				{}
		;

execute_param_clause: '(' expr_list ')'				{}
					| /* EMPTY */					{}
					;

/*****************************************************************************
 *
 *		QUERY:
 *				DEALLOCATE [PREPARE] <plan_name>
 *
 *****************************************************************************/

DeallocateStmt: DEALLOCATE name
					{}
				| DEALLOCATE PREPARE name
					{}
		;

/*****************************************************************************
 *
 *		QUERY:
 *				INSERT STATEMENTS
 *
 *****************************************************************************/

InsertStmt:
			INSERT INTO qualified_name insert_rest
				{}
		;

insert_rest:
			VALUES '(' insert_target_list ')'
				{}
			| DEFAULT VALUES
				{}
			| SelectStmt
				{}
			| '(' insert_column_list ')' VALUES '(' insert_target_list ')'
				{}
			| '(' insert_column_list ')' SelectStmt
				{}
		;

insert_column_list:
			insert_column_item						{}
			| insert_column_list ',' insert_column_item
					{}
		;

insert_column_item:
			ColId opt_indirection
				{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				DELETE STATEMENTS
 *
 *****************************************************************************/

DeleteStmt: DELETE_P FROM relation_expr where_clause
				{}
		;

LockStmt:	LOCK_P opt_table qualified_name_list opt_lock
				{}
		;

opt_lock:	IN_P lock_type MODE 			{}
			| /*EMPTY*/						{}
		;

lock_type:	ACCESS SHARE					{}
			| ROW SHARE						{}
			| ROW EXCLUSIVE					{}
			| SHARE UPDATE EXCLUSIVE		{}
			| SHARE							{}
			| SHARE ROW EXCLUSIVE			{}
			| EXCLUSIVE						{}
			| ACCESS EXCLUSIVE				{}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				UpdateStmt (UPDATE)
 *
 *****************************************************************************/

UpdateStmt: UPDATE relation_expr
			SET update_target_list
			from_clause
			where_clause
				{}
		|
		UPDATE relation_expr MODIFY BY func_name '(' opt_expr_list ')'
		{
			string membership = modifierFunctionCall (IFSQL_SCHEMA_NAME".make_bool(mship, nmship)", $5, $7);
			$$ = "update " + $2 + " set "
				"mship = "IFSQL_SCHEMA_NAME".get_mu (" + membership + "), "
				"nmship = "IFSQL_SCHEMA_NAME".get_nu (" + membership + ")";
		}
		;


/*****************************************************************************
 *
 *		QUERY:
 *				CURSOR STATEMENTS
 *
 *****************************************************************************/
DeclareCursorStmt: DECLARE name cursor_options CURSOR opt_hold FOR SelectStmt
				{}
		;

cursor_options: /*EMPTY*/					{}
			| cursor_options NO SCROLL		{}
			| cursor_options SCROLL			{}
			| cursor_options BINARY			{}
			| cursor_options INSENSITIVE	{}
		;

opt_hold: /* EMPTY */						{}
			| WITH HOLD						{}
			| WITHOUT HOLD					{}
		;

/*****************************************************************************
 *
 *		QUERY:
 *				SELECT STATEMENTS
 *
 *****************************************************************************/

/* A complete SELECT statement looks like this.
 *
 * The rule returns either a single SelectStmt node or a tree of them,
 * representing a set-operation tree.
 *
 * There is an ambiguity when a sub-SELECT is within an a_expr and there
 * are excess parentheses: do the parentheses belong to the sub-SELECT or
 * to the surrounding a_expr?  We don't really care, but yacc wants to know.
 * To resolve the ambiguity, we are careful to define the grammar so that
 * the decision is staved off as long as possible: as long as we can keep
 * absorbing parentheses into the sub-SELECT, we will do so, and only when
 * it's no longer possible to do that will we decide that parens belong to
 * the expression.	For example, in "SELECT (((SELECT 2)) + 3)" the extra
 * parentheses are treated as part of the sub-select.  The necessity of doing
 * it that way is shown by "SELECT (((SELECT 2)) UNION SELECT 2)".	Had we
 * parsed "((SELECT 2))" as an a_expr, it'd be too late to go back to the
 * SELECT viewpoint when we see the UNION.
 *
 * This approach is implemented by defining a nonterminal select_with_parens,
 * which represents a SELECT with at least one outer layer of parentheses,
 * and being careful to use select_with_parens, never '(' SelectStmt ')',
 * in the expression grammar.  We will then have shift-reduce conflicts
 * which we can resolve in favor of always treating '(' <select> ')' as
 * a select_with_parens.  To resolve the conflicts, the productions that
 * conflict with the select_with_parens productions are manually given
 * precedences lower than the precedence of ')', thereby ensuring that we
 * shift ')' (and then reduce to select_with_parens) rather than trying to
 * reduce the inner <select> nonterminal to something else.  We use UMINUS
 * precedence for this, which is a fairly arbitrary choice.
 *
 * To be able to define select_with_parens itself without ambiguity, we need
 * a nonterminal select_no_parens that represents a SELECT structure with no
 * outermost parentheses.  This is a little bit tedious, but it works.
 *
 * In non-expression contexts, we use SelectStmt which can represent a SELECT
 * with or without outer parentheses.
 */

SelectStmt: select_no_parens			%prec UMINUS
			| select_with_parens		%prec UMINUS
		;

select_with_parens:
			'(' select_no_parens ')'				{}
			| '(' select_with_parens ')'			{}
		;

/*
 *	FOR UPDATE may be before or after LIMIT/OFFSET.
 *	In <=7.2.X, LIMIT/OFFSET had to be after FOR UPDATE
 *	We now support both orderings, but prefer LIMIT/OFFSET before FOR UPDATE
 *	2002-08-28 bjm
 */
select_no_parens:
			simple_select						{}
			| select_clause sort_clause
				{}
			| select_clause opt_sort_clause for_update_clause opt_select_limit
				{}
			| select_clause opt_sort_clause select_limit opt_for_update_clause
				{}
		;

select_clause:
			simple_select							{}
			| select_with_parens					{}
		;

/*
 * This rule parses SELECT statements that can appear within set operations,
 * including UNION, INTERSECT and EXCEPT.  '(' and ')' can be used to specify
 * the ordering of the set operations.	Without '(' and ')' we want the
 * operations to be ordered per the precedence specs at the head of this file.
 *
 * As with select_no_parens, simple_select cannot have outer parentheses,
 * but can have parenthesized subclauses.
 *
 * Note that sort clauses cannot be included at this level --- SQL92 requires
 *		SELECT foo UNION SELECT bar ORDER BY baz
 * to be parsed as
 *		(SELECT foo UNION SELECT bar) ORDER BY baz
 * not
 *		SELECT foo UNION (SELECT bar ORDER BY baz)
 * Likewise FOR UPDATE and LIMIT.  Therefore, those clauses are described
 * as part of the select_no_parens production, not simple_select.
 * This does not limit functionality, because you can reintroduce sort and
 * limit clauses inside parentheses.
 *
 * NOTE: only the leftmost component SelectStmt should have INTO.
 * However, this is not checked by the grammar; parse analysis must check it.
 */
simple_select:
			SELECT opt_distinct select_recognizer target_list
			into_clause from_clause where_clause
			group_clause having_clause modify_clause
			{
				$$ = $1 + " " + $2 + getStmtVars().createTargetList() + " " +
					$5 + " " + $6 + " " + $7 + " " + $8 + " " + $9;
				if (getStmtVars().parent->setOperFlag)
					getStmtVars().parent->setOperFlag = false;
				else
					getStmtVars().parent->setOperAliasList = getStmtVars().selectAliasList;
				popStmtVars ();
			}
			
			| select_clause UNION opt_all set_operation_recognizer select_clause
			{
				string targetList, groupByList;
				getStmtVars().createUnionTargetList (targetList, groupByList);
				$$ = "select " + targetList + 
					" from (" + $1 + " UNION " + $5 + ") "IFSQL_SET_OPERATION_SUB_SELECT + 
						// UNION must be uppercase, otherwise sed will replace it with struct
					" group by " + groupByList;
			}
			
			| select_clause INTERSECT opt_all set_operation_recognizer select_clause
			{
				string targetList, groupByList;
				string aliasList = concatenateStrings (getStmtVars().setOperAliasList);
				getStmtVars().createIntersectTargetList (targetList, groupByList);
				$$ = "select " + targetList +
					" from ("
						"select 1 as "IFSQL_UNION_COUNT", " + aliasList +
						  " from (" + $1 + ") "IFSQL_SET_OPERATION_SUB_SELECT
						" UNION " // must be uppercase, otherwise sed will replace it with struct
						"select 2 as "IFSQL_UNION_COUNT", " + aliasList +
						  " from (" + $5 + ") "IFSQL_SET_OPERATION_SUB_SELECT
						") "IFSQL_SET_OPERATION_SUB_SELECT
					" group by " + groupByList +
					" having count ("IFSQL_UNION_COUNT") = 2";
			}
			
			| select_clause EXCEPT opt_all set_operation_recognizer select_clause
				
		;

select_recognizer:
			{ pushStmtVars (stSelect); }
		;

set_operation_recognizer:
			{ getStmtVars().setOperFlag = true; }
		;


into_clause:
			INTO OptTempTableName					{}
			| /*EMPTY*/								{}
		;

/*
 * Redundancy here is needed to avoid shift/reduce conflicts,
 * since TEMP is not a reserved word.  See also OptTemp.
 */
OptTempTableName:
			TEMPORARY opt_table qualified_name
				{}
			| TEMP opt_table qualified_name
				{}
			| LOCAL TEMPORARY opt_table qualified_name
				{}
			| LOCAL TEMP opt_table qualified_name
				{}
			| GLOBAL TEMPORARY opt_table qualified_name
				{}
			| GLOBAL TEMP opt_table qualified_name
				{}
			| TABLE qualified_name
				{}
			| qualified_name
				{}
		;

opt_table:	TABLE									{}
			| /*EMPTY*/								{}
		;

opt_all:	ALL										{}
			| DISTINCT								{}
			| /*EMPTY*/								{}
		;

/* We use (NIL) as a placeholder to indicate that all target expressions
 * should be placed in the DISTINCT list during parsetree analysis.
 */
opt_distinct:
			DISTINCT								{}
			| DISTINCT ON '(' expr_list ')'			{}
			| ALL									{}
			| /*EMPTY*/								{}
		;

opt_sort_clause:
			sort_clause								{}
			| /*EMPTY*/								{}
		;

sort_clause:
			ORDER BY sortby_list					{}
		;

sortby_list:
			sortby									{}
			| sortby_list ',' sortby				{}
		;

sortby:		a_expr USING qual_all_Op
				{}
			| a_expr ASC
				{}
			| a_expr DESC
				{}
			| a_expr
				{}
		;


select_limit:
			LIMIT select_limit_value OFFSET select_offset_value
				{}
			| OFFSET select_offset_value LIMIT select_limit_value
				{}
			| LIMIT select_limit_value
				{}
			| OFFSET select_offset_value
				{}
			| LIMIT select_limit_value ',' select_offset_value
				{}
		;

opt_select_limit:
			select_limit							{}
			| /* EMPTY */
					{}
		;

select_limit_value:
			a_expr									{}
			| ALL
				{}
		;

select_offset_value:
			a_expr									{}
		;

/*
 *	jimmy bell-style recursive queries aren't supported in the
 *	current system.
 *
 *	...however, recursive addattr and rename supported.  make special
 *	cases for these.
 */

group_clause:
			GROUP_P BY select_group_recognizer expr_list					{}
			| /*EMPTY*/								{}
		;

select_group_recognizer:
		{ getStmtVars().membership = 
			IFSQL_SCHEMA_NAME".make_bool ("
			  "avg ("IFSQL_SCHEMA_NAME".get_mu(" + getStmtVars().membership + ")), "
			  "avg ("IFSQL_SCHEMA_NAME".get_nu(" + getStmtVars().membership + "))"
			")";
		}
		;

having_clause:
			HAVING select_having_recognizer a_expr
			{ getStmtVars().stmtContext = scUnknown; }
			| /*EMPTY*/								{}
		;

select_having_recognizer:
			{ getStmtVars().stmtContext = scHaving; }
		;

modify_clause:
			MODIFY BY modifier_list
			| /*EMPTY*/ {}
		;

modifier_list:
			func_name '(' opt_expr_list ')'
			{ getStmtVars().modifyMembership ($1, $3); }
			
			| modifier_list ',' func_name '(' opt_expr_list ')'
			{ getStmtVars().modifyMembership ($3, $5); }
		;

for_update_clause:
			FOR UPDATE update_list					{}
			| FOR READ ONLY							{}
		;

opt_for_update_clause:
			for_update_clause						{}
			| /* EMPTY */							{}
		;

update_list:
			OF name_list							{}
			| /* EMPTY */							{}
		;

/*****************************************************************************
 *
 *	clauses common to all Optimizable Stmts:
 *		from_clause		- allow list of both JOIN expressions and table names
 *		where_clause	- qualifications for joins or restrictions
 *
 *****************************************************************************/

from_clause:
			FROM from_list	
				{ getStmtVars().createFromClauseMembership (); }
			| /*EMPTY*/								{}
		;

from_list:
			table_ref								{}
			| from_list ',' table_ref				{}
		;

/*
 * table_ref is where an alias clause can be attached.	Note we cannot make
 * alias_clause have an empty production because that causes parse conflicts
 * between table_ref := '(' joined_table ')' alias_clause
 * and joined_table := '(' joined_table ')'.  So, we must have the
 * redundant-looking productions here instead.
 */
table_ref:	relation_expr
				{ if (!isRelationCrisp ($1)) getStmtVars().fromList.push_back ($1); }
			| relation_expr alias_clause
				{ if (!isRelationCrisp ($1)) getStmtVars().fromList.push_back (getStmtVars().tableAlias); }
			| func_table
				{ getStmtVars().fromList.push_back ($1); }
			| func_table alias_clause
				{ getStmtVars().fromList.push_back (getStmtVars().tableAlias); }
			| func_table AS '(' TableFuncElementList ')'
				{ getStmtVars().fromList.push_back ($1); }
			| func_table AS ColId '(' TableFuncElementList ')'
				{ getStmtVars().fromList.push_back ($3); }
			| func_table ColId '(' TableFuncElementList ')'
				{ getStmtVars().fromList.push_back ($2); }
			| select_with_parens
				{}
			| select_with_parens alias_clause
				{ getStmtVars().fromList.push_back (getStmtVars().tableAlias); }
			| joined_table
				{}
			| '(' joined_table ')' alias_clause
				{ getStmtVars().fromList.push_back (getStmtVars().tableAlias); }
		;


/*
 * It may seem silly to separate joined_table from table_ref, but there is
 * method in SQL92's madness: if you don't do it this way you get reduce-
 * reduce conflicts, because it's not clear to the parser generator whether
 * to expect alias_clause after ')' or not.  For the same reason we must
 * treat 'JOIN' and 'join_type JOIN' separately, rather than allowing
 * join_type to expand to empty; if we try it, the parser generator can't
 * figure out when to reduce an empty join_type right after table_ref.
 *
 * Note that a CROSS JOIN is the same as an unqualified
 * INNER JOIN, and an INNER JOIN/ON has the same shape
 * but a qualification expression to limit membership.
 * A NATURAL JOIN implicitly matches column names between
 * tables and the shape is determined by which columns are
 * in common. We'll collect columns during the later transformations.
 */

joined_table:
			'(' joined_table ')'
				{}
			| table_ref CROSS JOIN table_ref
				{}
			| table_ref UNIONJOIN table_ref
				{}
			| table_ref join_type JOIN table_ref join_qual
				{}
			| table_ref JOIN table_ref join_qual
				{}
			| table_ref NATURAL join_type JOIN table_ref
				{}
			| table_ref NATURAL JOIN table_ref
				{}
		;

alias_clause:
			AS ColId '(' name_list ')'
				{ getStmtVars().tableAlias = $2; }
			| AS ColId
				{ getStmtVars().tableAlias = $2; }
			| ColId '(' name_list ')'
				{ getStmtVars().tableAlias = $1; }
			| ColId
				{ getStmtVars().tableAlias = $1; }
		;

join_type:	FULL join_outer							{}
			| LEFT join_outer						{}
			| RIGHT join_outer						{}
			| INNER_P								{}
		;

/* OUTER is just noise... */
join_outer: OUTER_P									{}
			| /*EMPTY*/								{}
		;

/* JOIN qualification clauses
 * Possibilities are:
 *	USING ( column list ) allows only unqualified column names,
 *						  which must match between tables.
 *	ON expr allows more general qualifications.
 *
 * We return USING as a List node, while an ON-expr will not be a List.
 */

join_qual:	USING '(' name_list ')'					{}
			| ON a_expr								{}
		;


relation_expr:
			qualified_name
				{}
			| qualified_name '*'
				{}
			| ONLY qualified_name
				{}
			| ONLY '(' qualified_name ')'
				{}
		;


func_table: func_name '(' ')'
				{}
			| func_name '(' expr_list ')'
				{}
		;


where_clause:
			WHERE select_where_recognizer a_expr							
			{
				if (getStmtVars().stmtType == stSelect)
				{
					if (getStmtVars().membership == "")
						getStmtVars().membership = IFSQL_SCHEMA_NAME".make_bool (" + $3 + ")";
					else
						getStmtVars().membership = IFSQL_SCHEMA_NAME".ifand (" +
							getStmtVars().membership + 
							", "IFSQL_SCHEMA_NAME".make_bool (" + $3 + "))";
					$$ = "where "IFSQL_SCHEMA_NAME".get_nu (" + getStmtVars().membership + ") < 1";
					getStmtVars().stmtContext = scUnknown;
				}
			}
			| /*EMPTY*/								{}
		;

select_where_recognizer:
			{ if (getStmtVars().stmtType == stSelect) getStmtVars().stmtContext = scWhere; }
		;

TableFuncElementList:
			TableFuncElement
				{}
			| TableFuncElementList ',' TableFuncElement
				{}
		;

TableFuncElement:	ColId Typename
				{}
		;

/*****************************************************************************
 *
 *	Type syntax
 *		SQL92 introduces a large amount of type-specific syntax.
 *		Define individual clauses to handle these cases, and use
 *		 the generic case to handle regular type-extensible Postgres syntax.
 *		- thomas 1997-10-10
 *
 *****************************************************************************/

Typename:	SimpleTypename opt_array_bounds
				{}
			| SETOF SimpleTypename opt_array_bounds
				{}
			| SimpleTypename ARRAY '[' Iconst ']'
				{}
			| SETOF SimpleTypename ARRAY '[' Iconst ']'
				{}
		;

opt_array_bounds:
			opt_array_bounds '[' ']'
					{}
			| opt_array_bounds '[' Iconst ']'
					{}
			| /*EMPTY*/
					{}
		;

/*
 * XXX ideally, the production for a qualified typename should be ColId attrs
 * (there's no obvious reason why the first name should need to be restricted)
 * and should be an alternative of GenericType (so that it can be used to
 * specify a type for a literal in AExprConst).  However doing either causes
 * reduce/reduce conflicts that I haven't been able to find a workaround
 * for.  FIXME later.
 */
SimpleTypename:
			GenericType								{}
			| Numeric								{}
			| Bit									{}
			| Character								{}
			| ConstDatetime							{}
			| ConstInterval opt_interval
				{}
			| ConstInterval '(' Iconst ')' opt_interval
				{}
			| type_name attrs
				{}
		;

/* We have a separate ConstTypename to allow defaulting fixed-length
 * types such as CHAR() and BIT() to an unspecified length.
 * SQL9x requires that these default to a length of one, but this
 * makes no sense for constructs like CHAR 'hi' and BIT '0101',
 * where there is an obvious better choice to make.
 * Note that ConstInterval is not included here since it must
 * be pushed up higher in the rules to accomodate the postfix
 * options (e.g. INTERVAL '1' YEAR).
 */
ConstTypename:
			GenericType								{}
			| Numeric								{}
			| ConstBit								{}
			| ConstCharacter						{}
			| ConstDatetime							{}
		;

GenericType:
			type_name
				{}
		;

/* SQL92 numeric data types
 * Check FLOAT() precision limits assuming IEEE floating types.
 * - thomas 1997-09-18
 * Provide real DECIMAL() and NUMERIC() implementations now - Jan 1998-12-30
 */
Numeric:	INT_P
				{}
			| INTEGER
				{}
			| SMALLINT
				{}
			| BIGINT
				{}
			| REAL
				{}
			| FLOAT_P opt_float
				{}
			| DOUBLE_P PRECISION
				{}
			| DECIMAL_P opt_decimal
				{}
			| DEC opt_decimal
				{}
			| NUMERIC opt_numeric
				{}
			| BOOLEAN_P
				{}
		;

opt_float:	'(' Iconst ')'
				{}
			| /*EMPTY*/
				{}
		;

opt_numeric:
			'(' Iconst ',' Iconst ')'
				{}
			| '(' Iconst ')'
				{}
			| /*EMPTY*/
				{}
		;

opt_decimal:
			'(' Iconst ',' Iconst ')'
				{}
			| '(' Iconst ')'
				{}
			| /*EMPTY*/
				{}
		;


/*
 * SQL92 bit-field data types
 * The following implements BIT() and BIT VARYING().
 */
Bit:		BitWithLength
				{}
			| BitWithoutLength
				{}
		;

/* ConstBit is like Bit except "BIT" defaults to unspecified length */
/* See notes for ConstCharacter, which addresses same issue for "CHAR" */
ConstBit:	BitWithLength
				{}
			| BitWithoutLength
				{}
		;

BitWithLength:
			BIT opt_varying '(' Iconst ')'
				{}
		;

BitWithoutLength:
			BIT opt_varying
				{}
		;


/*
 * SQL92 character data types
 * The following implements CHAR() and VARCHAR().
 */
Character:  CharacterWithLength
				{}
			| CharacterWithoutLength
				{}
		;

ConstCharacter:  CharacterWithLength
				{}
			| CharacterWithoutLength
				{}
		;

CharacterWithLength:  character '(' Iconst ')' opt_charset
				{}
		;

CharacterWithoutLength:	 character opt_charset
				{}
		;

character:	CHARACTER opt_varying
										{}
			| CHAR_P opt_varying
										{}
			| VARCHAR
										{}
			| NATIONAL CHARACTER opt_varying
										{}
			| NATIONAL CHAR_P opt_varying
										{}
			| NCHAR opt_varying
										{}
		;

opt_varying:
			VARYING									{}
			| /*EMPTY*/								{}
		;

opt_charset:
			CHARACTER SET ColId						{}
			| /*EMPTY*/								{}
		;

ConstDatetime:
			TIMESTAMP '(' Iconst ')' opt_timezone
				{}
			| TIMESTAMP opt_timezone
				{}
			| TIME '(' Iconst ')' opt_timezone
				{}
			| TIME opt_timezone
				{}
		;

ConstInterval:
			INTERVAL								{}
		;

opt_timezone:
			WITH TIME ZONE							{}
			| WITHOUT TIME ZONE						{}
			| /*EMPTY*/								{}
		;

opt_interval:
			YEAR_P									{}
			| MONTH_P								{}
			| DAY_P									{}
			| HOUR_P								{}
			| MINUTE_P								{}
			| SECOND_P								{}
			| YEAR_P TO MONTH_P
					{}
			| DAY_P TO HOUR_P
					{}
			| DAY_P TO MINUTE_P
					{}
			| DAY_P TO SECOND_P
					{}
			| HOUR_P TO MINUTE_P
					{}
			| HOUR_P TO SECOND_P
					{}
			| MINUTE_P TO SECOND_P
					{}
			| /*EMPTY*/								{}
		;


/*****************************************************************************
 *
 *	expression grammar
 *
 *****************************************************************************/

/* Expressions using row descriptors
 * Define row_descriptor to allow yacc to break the reduce/reduce conflict
 * with singleton expressions. Use SQL99's ROW keyword to allow rows of
 * one element.
 */
r_expr:  row IN_P select_with_parens
				{}
			| row NOT IN_P select_with_parens
				{}
			| row qual_all_Op sub_type select_with_parens
			%prec Op
				{}
			| row qual_all_Op select_with_parens
			%prec Op
				{}
			| row qual_all_Op row
			%prec Op
				{}
			| row IS NULL_P
				{}
			| row IS NOT NULL_P
				{}
			| row OVERLAPS row
				{}
			| row IS DISTINCT FROM row
			%prec IS
				{}
		;

/* Explicit row production.
 * SQL99 allows an optional ROW keyword, so we can now do single-element productions
 * without conflicting with the parenthesized a_expr production.
 */
row:  ROW '(' row_descriptor ')'					{}
			| ROW '(' a_expr ')'					{}
			| ROW '(' ')'							{}
			| '(' row_descriptor ')'				{}
		;

row_descriptor:  expr_list ',' a_expr				{}
		;

sub_type:	ANY										{}
			| SOME									{}
			| ALL									{}
		;

all_Op:		Op										{}
			| MathOp								{}
		;

MathOp:		 '+'									{}
			| '-'									{}
			| '*'									{}
			| '/'									{}
			| '%'									{}
			| '^'									{}
			| '<'									{}
			| '>'									{}
			| '='									{}
		;

qual_Op:	Op
					{}
			| OPERATOR '(' any_operator ')'			{}
		;

qual_all_Op:
			all_Op
					{}
			| OPERATOR '(' any_operator ')'			{}
		;

/*
 * General expressions
 * This is the heart of the expression syntax.
 *
 * We have two expression types: a_expr is the unrestricted kind, and
 * b_expr is a subset that must be used in some places to avoid shift/reduce
 * conflicts.  For example, we can't do BETWEEN as "BETWEEN a_expr AND a_expr"
 * because that use of AND conflicts with AND as a boolean operator.  So,
 * b_expr is used in BETWEEN and we remove boolean keywords from b_expr.
 *
 * Note that '(' a_expr ')' is a b_expr, so an unrestricted expression can
 * always be used by surrounding it with parens.
 *
 * c_expr is all the productions that are common to a_expr and b_expr;
 * it's factored out just to eliminate redundant coding.
 */
a_expr:		c_expr									{}
			| a_expr TYPECAST Typename
					{}
			| a_expr AT TIME ZONE c_expr
				{}
		/*
		 * These operators must be called out explicitly in order to make use
		 * of yacc/bison's automatic operator-precedence handling.  All other
		 * operator names are handled by the generic productions using "Op",
		 * below; and all those operators will have the same precedence.
		 *
		 * If you add more explicitly-known operators, be sure to add them
		 * also to b_expr and to the MathOp list above.
		 */
			| '+' a_expr					%prec UMINUS
				{}
			| '-' a_expr					%prec UMINUS
				{}
			| '%' a_expr
				{}
			| '^' a_expr
				{}
			| a_expr '%'
				{}
			| a_expr '^'
				{}
			| a_expr '+' a_expr
				{}
			| a_expr '-' a_expr
				{}
			| a_expr '*' a_expr
				{}
			| a_expr '/' a_expr
				{}
			| a_expr '%' a_expr
				{}
			| a_expr '^' a_expr
				{}
			| a_expr '<' a_expr
				{}
			| a_expr '>' a_expr
				{}
			| a_expr '=' a_expr
				{}

			| a_expr qual_Op a_expr				%prec Op
				{}
			| qual_Op a_expr					%prec Op
				{}
			| a_expr qual_Op					%prec POSTFIXOP
				{}

			| a_expr AND a_expr
			{
				if (getStmtVars().stmtType == stSelect && getStmtVars().stmtContext == scWhere)
					$$ = IFSQL_SCHEMA_NAME".ifand ("
						IFSQL_SCHEMA_NAME".make_bool (" + $1 + "), "
						IFSQL_SCHEMA_NAME".make_bool (" + $3 + "))"; 
			}
			
			| a_expr OR a_expr
			{
				if (getStmtVars().stmtType == stSelect && getStmtVars().stmtContext == scWhere)
					$$ = IFSQL_SCHEMA_NAME".ifor ("
						IFSQL_SCHEMA_NAME".make_bool (" + $1 + "), "
						IFSQL_SCHEMA_NAME".make_bool (" + $3 + "))"; 
			}
			
			| NOT a_expr
			{
				if (getStmtVars().stmtType == stSelect && getStmtVars().stmtContext == scWhere)
					$$ = IFSQL_SCHEMA_NAME".ifnot ("IFSQL_SCHEMA_NAME".make_bool (" + $2 + "))";
			}

			| a_expr LIKE a_expr
				{}
			| a_expr LIKE a_expr ESCAPE a_expr
				{}
			| a_expr NOT LIKE a_expr
				{}
			| a_expr NOT LIKE a_expr ESCAPE a_expr
				{}
			| a_expr ILIKE a_expr
				{}
			| a_expr ILIKE a_expr ESCAPE a_expr
				{}
			| a_expr NOT ILIKE a_expr
				{}
			| a_expr NOT ILIKE a_expr ESCAPE a_expr
				{}

			| a_expr SIMILAR TO a_expr				%prec SIMILAR
				{}
			| a_expr SIMILAR TO a_expr ESCAPE a_expr
				{}
			| a_expr NOT SIMILAR TO a_expr			%prec SIMILAR
				{}
			| a_expr NOT SIMILAR TO a_expr ESCAPE a_expr
				{}

			/* NullTest clause
			 * Define SQL92-style Null test clause.
			 * Allow two forms described in the standard:
			 *	a IS NULL
			 *	a IS NOT NULL
			 * Allow two SQL extensions
			 *	a ISNULL
			 *	a NOTNULL
			 */
			| a_expr ISNULL
				{}
			| a_expr IS NULL_P
				{}
			| a_expr NOTNULL
				{}
			| a_expr IS NOT NULL_P
				{}
			| a_expr IS TRUE_P
				{}
			| a_expr IS NOT TRUE_P
				{}
			| a_expr IS FALSE_P
				{}
			| a_expr IS NOT FALSE_P
				{}
			| a_expr IS UNKNOWN
				{}
			| a_expr IS NOT UNKNOWN
				{}
			| a_expr IS DISTINCT FROM a_expr			%prec IS
				{}
			| a_expr IS OF '(' type_list ')'			%prec IS
				{}
			| a_expr IS NOT OF '(' type_list ')'		%prec IS
				{}
			| a_expr BETWEEN b_expr AND b_expr			%prec BETWEEN
				{}
			| a_expr NOT BETWEEN b_expr AND b_expr		%prec BETWEEN
				{}
			| a_expr IN_P in_expr
				{}
			| a_expr NOT IN_P in_expr
				{}
			| a_expr qual_all_Op sub_type select_with_parens %prec Op
				{}
			| a_expr qual_all_Op sub_type '(' a_expr ')' %prec Op
				{}
			| UNIQUE select_with_parens %prec Op
				{}
			| r_expr
				{}
		;

/*
 * Restricted expressions
 *
 * b_expr is a subset of the complete expression syntax defined by a_expr.
 *
 * Presently, AND, NOT, IS, and IN are the a_expr keywords that would
 * cause trouble in the places where b_expr is used.  For simplicity, we
 * just eliminate all the boolean-keyword-operator productions from b_expr.
 */
b_expr:		c_expr
				{}
			| b_expr TYPECAST Typename
				{}
			| '+' b_expr					%prec UMINUS
				{}
			| '-' b_expr					%prec UMINUS
				{}
			| '%' b_expr
				{}
			| '^' b_expr
				{}
			| b_expr '%'
				{}
			| b_expr '^'
				{}
			| b_expr '+' b_expr
				{}
			| b_expr '-' b_expr
				{}
			| b_expr '*' b_expr
				{}
			| b_expr '/' b_expr
				{}
			| b_expr '%' b_expr
				{}
			| b_expr '^' b_expr
				{}
			| b_expr '<' b_expr
				{}
			| b_expr '>' b_expr
				{}
			| b_expr '=' b_expr
				{}
			| b_expr qual_Op b_expr				%prec Op
				{}
			| qual_Op b_expr					%prec Op
				{}
			| b_expr qual_Op					%prec POSTFIXOP
				{}
			| b_expr IS DISTINCT FROM b_expr	%prec IS
				{}
			| b_expr IS OF '(' type_list ')'	%prec IS
				{}
			| b_expr IS NOT OF '(' type_list ')'	%prec IS
				{}
		;

/*
 * Productions that can be used in both a_expr and b_expr.
 *
 * Note: productions that refer recursively to a_expr or b_expr mostly
 * cannot appear here.	However, it's OK to refer to a_exprs that occur
 * inside parentheses, such as function arguments; that cannot introduce
 * ambiguity to the b_expr syntax.
 */
c_expr:		
			MSHIP
				{ 
					if (getStmtVars().stmtType == stSelect && getStmtVars().stmtContext == scHaving)
						$$ = IFSQL_SCHEMA_NAME".get_mu (" + getStmtVars().membership + ")";
				}
			| NMSHIP
				{ 
					if (getStmtVars().stmtType == stSelect && getStmtVars().stmtContext == scHaving)
						$$ = IFSQL_SCHEMA_NAME".get_nu (" + getStmtVars().membership + ")";
				}
			
			| columnref								
			| AexprConst							{}
			| PARAM attrs opt_indirection
				{}
			| '(' a_expr ')' attrs opt_indirection
				{}
			| '(' a_expr ')' opt_indirection
				{}
			| case_expr
				{}
			| func_name '(' ')'
				{}
			| func_name '(' expr_list ')'
				{}
			| func_name '(' ALL expr_list ')'
				{}
			| func_name '(' DISTINCT expr_list ')'
				{}
			| func_name '(' '*' ')'
				{}
			| CURRENT_DATE
				{}
			| CURRENT_TIME
				{}
			| CURRENT_TIME '(' Iconst ')'
				{}
			| CURRENT_TIMESTAMP
				{}
			| CURRENT_TIMESTAMP '(' Iconst ')'
				{}
			| LOCALTIME
				{}
			| LOCALTIME '(' Iconst ')'
				{}
			| LOCALTIMESTAMP
				{}
			| LOCALTIMESTAMP '(' Iconst ')'
				{}
			| CURRENT_USER
				{}
			| SESSION_USER
				{}
			| USER
				{}
			| CAST '(' a_expr AS Typename ')'
				{}
			| EXTRACT '(' extract_list ')'
				{}
			| OVERLAY '(' overlay_list ')'
				{}
			| POSITION '(' position_list ')'
				{}
			| SUBSTRING '(' substr_list ')'
				{}
			| TREAT '(' a_expr AS Typename ')'
				{}
			| TRIM '(' BOTH trim_list ')'
				{}
			| TRIM '(' LEADING trim_list ')'
				{}
			| TRIM '(' TRAILING trim_list ')'
				{}
			| TRIM '(' trim_list ')'
				{}
			| CONVERT '(' a_expr USING any_name ')'
				{}
			| CONVERT '(' expr_list ')'
				{}
			| select_with_parens			%prec UMINUS
				{}
			| EXISTS select_with_parens
				{}
			| ARRAY select_with_parens
				{}
			| ARRAY array_expr
				{}
		;

/*
 * Supporting nonterminals for expressions.
 */

opt_indirection:
			opt_indirection '[' a_expr ']'
				{}
			| opt_indirection '[' a_expr ':' a_expr ']'
				{}
			| /*EMPTY*/
				{}
		;

expr_list:	a_expr
				{}
			| expr_list ',' a_expr
				{}
		;

opt_expr_list:	expr_list {}
		|
		/* EMPTY */ {}
		;

extract_list:
			extract_arg FROM a_expr
				{}
			| /*EMPTY*/								{}
		;

type_list:  type_list ',' Typename
				{}
			| Typename
				{}
		;

array_expr_list: array_expr
				{}
			| array_expr_list ',' array_expr
				{}
		;

array_expr: '[' expr_list ']'
				{}
			| '[' array_expr_list ']'
				{}
		;

/* Allow delimited string SCONST in extract_arg as an SQL extension.
 * - thomas 2001-04-12
 */

extract_arg:
			IDENT									{}
			| YEAR_P								{}
			| MONTH_P								{}
			| DAY_P									{}
			| HOUR_P								{}
			| MINUTE_P								{}
			| SECOND_P								{}
			| SCONST								{}
		;

/* OVERLAY() arguments
 * SQL99 defines the OVERLAY() function:
 * o overlay(text placing text from int for int)
 * o overlay(text placing text from int)
 */
overlay_list:
			a_expr overlay_placing substr_from substr_for
				{}
			| a_expr overlay_placing substr_from
				{}
		;

overlay_placing:
			PLACING a_expr
				{}
		;

/* position_list uses b_expr not a_expr to avoid conflict with general IN */

position_list:
			b_expr IN_P b_expr						{}
			| /*EMPTY*/								{}
		;

/* SUBSTRING() arguments
 * SQL9x defines a specific syntax for arguments to SUBSTRING():
 * o substring(text from int for int)
 * o substring(text from int) get entire string from starting point "int"
 * o substring(text from pattern) get entire string matching pattern
 * o substring(text for int) get first "int" characters of string
 * We also want to implement generic substring functions which accept
 * the usual generic list of arguments. So we will accept both styles
 * here, and convert the SQL9x style to the generic list for further
 * processing. - thomas 2000-11-28
 */
substr_list:
			a_expr substr_from substr_for
				{}
			| a_expr substr_for substr_from
				{}
			| a_expr substr_from
				{}
			| a_expr substr_for
				{}
			| expr_list
				{}
			| /*EMPTY*/
				{}
		;

substr_from:
			FROM a_expr								{}
		;

substr_for: FOR a_expr								{}
		;

trim_list:	a_expr FROM expr_list					{}
			| FROM expr_list						{}
			| expr_list								{}
		;

in_expr:	select_with_parens
				{}
			| '(' expr_list ')'						{}
		;

/* Case clause
 * Define SQL92-style case clause.
 * Allow all four forms described in the standard:
 * - Full specification
 *	CASE WHEN a = b THEN c ... ELSE d END
 * - Implicit argument
 *	CASE a WHEN b THEN c ... ELSE d END
 * - Conditional NULL
 *	NULLIF(x,y)
 *	same as CASE WHEN x = y THEN NULL ELSE x END
 * - Conditional substitution from list, use first non-null argument
 *	COALESCE(a,b,...)
 * same as CASE WHEN a IS NOT NULL THEN a WHEN b IS NOT NULL THEN b ... END
 * - thomas 1998-11-09
 * 
 * NULLIF and COALESCE have become first class nodes to
 * prevent double evaluation of arguments.
 * - Kris Jurka 2003-02-11
 */
case_expr:	CASE case_arg when_clause_list case_default END_P
				{}
			| NULLIF '(' a_expr ',' a_expr ')'
				{}
			| COALESCE '(' expr_list ')'
				{}
		;

when_clause_list:
			/* There must be at least one */
			when_clause								{}
			| when_clause_list when_clause			{}
		;

when_clause:
			WHEN a_expr THEN a_expr
				{}
		;

case_default:
			ELSE a_expr								{}
			| /*EMPTY*/								{}
		;

case_arg:	a_expr									{}
			| /*EMPTY*/								{}
		;

/*
 * columnref starts with relation_name not ColId, so that OLD and NEW
 * references can be accepted.	Note that when there are more than two
 * dotted names, the first name is not actually a relation name...
 */
columnref:	relation_name opt_indirection
				{}
			| dotted_name opt_indirection
				{}
		;

dotted_name:
			relation_name attrs
					{}
		;

attrs:		'.' attr_name
					{}
			| '.' '*'
					{}
			| '.' attr_name attrs
					{}
		;


/*****************************************************************************
 *
 *	target lists
 *
 *****************************************************************************/

/* Target lists as found in SELECT ... and INSERT VALUES ( ... ) */

target_list:
			select_target_el
				{ }
			| target_list ',' select_target_el
				{ }
		;

select_target_el:
			a_expr AS ColLabel
				{ getStmtVars().selectAliasList.push_back ($3); getStmtVars().selectList.push_back ($$); }
			| a_expr
				{ getStmtVars().selectAliasList.push_back ($$); getStmtVars().selectList.push_back ($$); }
			| '*'
				{ getStmtVars().selectAliasList.push_back ($$); getStmtVars().selectList.push_back ($$); }
		;

/* AS is not optional because shift/red conflict with unary ops */
target_el:	a_expr AS ColLabel
				{}
			| a_expr
				{}
			| '*'
				{}
		;

/* Target list as found in UPDATE table SET ...
| '(' row_ ')' = '(' row_ ')'
{}
 */
update_target_list:
			update_target_el						{}
			| update_target_list ',' update_target_el {}
		;

update_target_el:
			ColId opt_indirection '=' a_expr
				{}
			| ColId opt_indirection '=' DEFAULT
				{}

		;

insert_target_list:
			insert_target_el						{}
			| insert_target_list ',' insert_target_el {}
		;

insert_target_el:
			target_el								{}
			| DEFAULT
				{}
		;


/*****************************************************************************
 *
 *	Names and constants
 *
 *****************************************************************************/

relation_name:
			SpecialRuleRelation						{}
			| ColId									{}
		;

qualified_name_list:
			qualified_name							{}
			| qualified_name_list ',' qualified_name {}
		;

qualified_name:
			relation_name
				{}
			| dotted_name
				{}
		;

name_list:	name
					{}
			| name_list ',' name
					{}
		;


name:		ColId									{};

database_name:
			ColId									{};

access_method:
			ColId									{};

attr_name:	ColId									{};

index_name: ColId									{};

file_name:	Sconst									{};

func_name:	function_name
					{}
			| dotted_name							{}
		;


/*
 * Constants
 */
AexprConst: Iconst
				{}
			| FCONST
				{}
			| Sconst
				{}
			| BCONST
				{}
			| XCONST
				{}
			| ConstTypename Sconst
				{}
			| ConstInterval Sconst opt_interval
				{}
			| ConstInterval '(' Iconst ')' Sconst opt_interval
				{}
			| PARAM opt_indirection
				{}
			| TRUE_P
				{}
			| FALSE_P
				{}
			| NULL_P
				{}
		;

Iconst:		ICONST									{};
Sconst:		SCONST									{};
UserId:		ColId									{};

/*
 * Name classification hierarchy.
 *
 * IDENT is the lexeme returned by the lexer for identifiers that match
 * no known keyword.  In most cases, we can accept certain keywords as
 * names, not only IDENTs.	We prefer to accept as many such keywords
 * as possible to minimize the impact of "reserved words" on programmers.
 * So, we divide names into several possible classes.  The classification
 * is chosen in part to make keywords acceptable as names wherever possible.
 */

/* Column identifier --- names that can be column, table, etc names.
 */
ColId:		IDENT									{}
			| unreserved_keyword					{}
			| col_name_keyword						{}
		;

/* Type identifier --- names that can be type names.
 */
type_name:	IDENT									{}
			| unreserved_keyword					{}
		;

/* Function identifier --- names that can be function names.
 */
function_name:
			IDENT									{}
			| unreserved_keyword					{}
			| func_name_keyword						{}
		;

/* Column label --- allowed labels in "AS" clauses.
 * This presently includes *all* Postgres keywords.
 */
ColLabel:	IDENT									{}
			| unreserved_keyword					{}
			| col_name_keyword						{}
			| func_name_keyword						{}
			| reserved_keyword						{}
		;


/*
 * Keyword classification lists.  Generally, every keyword present in
 * the Postgres grammar should appear in exactly one of these lists.
 *
 * Put a new keyword into the first list that it can go into without causing
 * shift or reduce conflicts.  The earlier lists define "less reserved"
 * categories of keywords.
 */

/* "Unreserved" keywords --- available for use as any kind of name.
 */
unreserved_keyword:
			  ABORT_P
			| ABSOLUTE_P
			| ACCESS
			| ACTION
			| ADD
			| AFTER
			| AGGREGATE
			| ALTER
			| ASSERTION
			| ASSIGNMENT
			| AT
			| BACKWARD
			| BEFORE
			| BEGIN_P
			| BY
			| CACHE
			| CALLED
			| CASCADE
			| CHAIN
			| CHARACTERISTICS
			| CHECKPOINT
			| CLASS
			| CLOSE
			| CLUSTER
			| COMMENT
			| COMMIT
			| COMMITTED
			| CONSTRAINTS
			| CONVERSION_P
			| COPY
			| CREATEDB
			| CREATEUSER
			| CURSOR
			| CYCLE
			| DATABASE
			| DAY_P
			| DEALLOCATE
			| DECLARE
			| DEFAULTS
			| DEFERRED
			| DEFINER
			| DELETE_P
			| DELIMITER
			| DELIMITERS
			| DOMAIN_P
			| DOUBLE_P
			| DROP
			| EACH
			| ENCODING
			| ENCRYPTED
			| ESCAPE
			| EXCLUDING
			| EXCLUSIVE
			| EXECUTE
			| EXPLAIN
			| EXTERNAL
			| FETCH
			| FIRST_P
			| FORCE
			| FORWARD
			| FUNCTION
			| GLOBAL
			| HANDLER
			| HOLD
			| HOUR_P
			| IMMEDIATE
			| IMMUTABLE
			| IMPLICIT_P
			| INCLUDING
			| INCREMENT
			| INDEX
			| INHERITS
			| INOUT
			| INPUT_P
			| INSENSITIVE
			| INSERT
			| INSTEAD
			| INVOKER
			| ISOLATION
			| KEY
			| LANCOMPILER
			| LANGUAGE
			| LAST_P
			| LEVEL
			| LISTEN
			| LOAD
			| LOCAL
			| LOCATION
			| LOCK_P
			| MATCH
			| MAXVALUE
			| MINUTE_P
			| MINVALUE
			| MODE
			| MONTH_P
			| MOVE
			| NAMES
			| NATIONAL
			| NEXT
			| NO
			| NOCREATEDB
			| NOCREATEUSER
			| NOTHING
			| NOTIFY
			| OF
			| OIDS
			| OPERATOR
			| OPTION
			| OUT_P
			| OWNER
			| PARTIAL
			| PASSWORD
			| PATH_P
			| PENDANT
			| PRECISION
			| PREPARE
			| PRESERVE
			| PRIOR
			| PRIVILEGES
			| PROCEDURAL
			| PROCEDURE
			| READ
			| RECHECK
			| REINDEX
			| RELATIVE_P
			| RENAME
			| REPLACE
			| RESET
			| RESTART
			| RESTRICT
			| RETURNS
			| REVOKE
			| ROLLBACK
			| ROWS
			| RULE
			| SCHEMA
			| SCROLL
			| SECOND_P
			| SECURITY
			| SEQUENCE
			| SERIALIZABLE
			| SESSION
			| SET
			| SHARE
			| SHOW
			| SIMPLE
			| STABLE
			| START
			| STATEMENT
			| STATISTICS
			| STDIN
			| STDOUT
			| STORAGE
			| SYSID
			| STRICT_P
			| TEMP
			| TEMPLATE
			| TEMPORARY
			| TOAST
			| TRANSACTION
			| TRIGGER
			| TRUNCATE
			| TRUSTED
			| TYPE_P
			| UNENCRYPTED
			| UNKNOWN
			| UNLISTEN
			| UNTIL
			| UPDATE
			| USAGE
			| VACUUM
			| VALID
			| VALIDATOR
			| VALUES
			| VARYING
			| VERSION
			| VIEW
			| VOLATILE
			| WITH
			| WITHOUT
			| WORK
			| WRITE
			| YEAR_P
			| ZONE
		;

/* Column identifier --- keywords that can be column, table, etc names.
 *
 * Many of these keywords will in fact be recognized as type or function
 * names too; but they have special productions for the purpose, and so
 * can't be treated as "generic" type or function names.
 *
 * The type names appearing here are not usable as function names
 * because they can be followed by '(' in typename productions, which
 * looks too much like a function call for an LR(1) parser.
 */
col_name_keyword:
			  BIGINT
			| BIT
			| BOOLEAN_P
			| CHAR_P
			| CHARACTER
			| COALESCE
			| CONVERT
			| DEC
			| DECIMAL_P
			| EXISTS
			| EXTRACT
			| FLOAT_P
			| INT_P
			| INTEGER
			| INTERVAL
			| NCHAR
			| NONE
			| NULLIF
			| NUMERIC
			| OVERLAY
			| POSITION
			| REAL
			| ROW
			| SETOF
			| SMALLINT
			| SUBSTRING
			| TIME
			| TIMESTAMP
			| TREAT
			| TRIM
			| VARCHAR
		;

/* Function identifier --- keywords that can be function names.
 *
 * Most of these are keywords that are used as operators in expressions;
 * in general such keywords can't be column names because they would be
 * ambiguous with variables, but they are unambiguous as function identifiers.
 *
 * Do not include POSITION, SUBSTRING, etc here since they have explicit
 * productions in a_expr to support the goofy SQL9x argument syntax.
 * - thomas 2000-11-28
 */
func_name_keyword:
			  AUTHORIZATION
			| BETWEEN
			| BINARY
			| CROSS
			| FREEZE
			| FULL
			| ILIKE
			| IN_P
			| INNER_P
			| IS
			| ISNULL
			| JOIN
			| LEFT
			| LIKE
			| NATURAL
			| NOTNULL
			| OUTER_P
			| OVERLAPS
			| RIGHT
			| SIMILAR
			| VERBOSE
		;

/* Reserved keyword --- these keywords are usable only as a ColLabel.
 *
 * Keywords appear here if they could not be distinguished from variable,
 * type, or function names in some contexts.  Don't put things here unless
 * forced to.
 */
reserved_keyword:
			  ALL
			| ANALYSE
			| ANALYZE
			| AND
			| ANY
			| ARRAY
			| AS
			| ASC
			| BOTH
			| CASE
			| CAST
			| CHECK
			| COLLATE
			| COLUMN
			| CONSTRAINT
			| CREATE
			| CURRENT_DATE
			| CURRENT_TIME
			| CURRENT_TIMESTAMP
			| CURRENT_USER
			| DEFAULT
			| DEFERRABLE
			| DESC
			| DISTINCT
			| DO
			| ELSE
			| END_P
			| EXCEPT
			| FALSE_P
			| FOR
			| FOREIGN
			| FROM
			| GRANT
			| GROUP_P
			| HAVING
			| INITIALLY
			| INTERSECT
			| INTO
			| LEADING
			| LIMIT
			| LOCALTIME
			| LOCALTIMESTAMP
			| NEW
			| NOT
			| NULL_P
			| OFF
			| OFFSET
			| OLD
			| ON
			| ONLY
			| OR
			| ORDER
			| PLACING
			| PRIMARY
			| REFERENCES
			| SELECT
			| SESSION_USER
			| SOME
			| TABLE
			| THEN
			| TO
			| TRAILING
			| TRUE_P
			| UNION
			| UNIQUE
			| USER
			| USING
			| WHEN
			| WHERE
		;


SpecialRuleRelation:
			OLD
				{}
			| NEW
				{}
		;

%%

void StmtVars::createFromClauseMembership ()
{
	while (!fromList.empty())
	{
		string relationMembership = IFSQL_SCHEMA_NAME".make_bool (coalesce (" + fromList.front() + ".mship, 0), coalesce (" + fromList.front() + ".nmship, 1))";
		if (membership == "")
			membership = relationMembership;
		else
			membership = IFSQL_SCHEMA_NAME".ifand (" + membership + ", " + relationMembership + ")";
		fromList.pop_front();
	}
}

void StmtVars::modifyMembership (const string& modifierName, const string& modifierParams)
{
	membership = modifierFunctionCall (membership, modifierName, modifierParams);
}

string StmtVars::createTargetList ()
{
	string result;
	while (!selectList.empty())
	{
		if (result != "")
			result += ", ";
		if (selectList.front() == "mship")
			result += IFSQL_SCHEMA_NAME".get_mu (" + membership + ") as mship";
		else if (selectList.front() == "nmship")
			result += IFSQL_SCHEMA_NAME".get_nu (" + membership + ") as nmship";
		else
			result += selectList.front();
		selectList.pop_front();
	}
	return result;
}

static void createSetOperTargetList (const deque <string>& aliasList, string& targetList, string& groupByList, bool isUnion)
{
	targetList = groupByList = "";
	for (deque <string> ::const_iterator it = aliasList.begin(); it != aliasList.end(); ++it)
	{
		bool forGroupByList = false;
		if (targetList != "")
			targetList += ", ";
		if (*it == "mship")
			targetList += string (isUnion ? "max" : "min") + " (mship) as mship";
		else if (*it == "nmship")
			targetList += string (isUnion ? "min" : "max") + " (nmship) as nmship";
		else
		{
			targetList += *it;
			forGroupByList = true;
		}
		if (forGroupByList)
		{
			if (groupByList != "")
				groupByList += ", ";
			groupByList += *it;
		}
	}
}

void StmtVars::createUnionTargetList (string& targetList, string& groupByList)
{
	createSetOperTargetList (setOperAliasList, targetList, groupByList, true);
}

void StmtVars::createIntersectTargetList (string& targetList, string& groupByList)
{
	createSetOperTargetList (setOperAliasList, targetList, groupByList, false);
}

static void pushStmtVars (StmtType stmtType)
{
	StmtVars stmtVars;
	stmtVars.stmtType = stmtType;
	stmtVars.stmtContext = scUnknown;
	stmtVars.setOperFlag = false;
	
	if (stmtVarsStack.empty()) stmtVars.parent = 0;
	else stmtVars.parent = &stmtVarsStack.back();
	
	stmtVarsStack.push_back (stmtVars);
}

static void popStmtVars ()
{
	stmtVarsStack.pop_back ();
}

static StmtVars& getStmtVars ()
{
	return stmtVarsStack.back();
}

static string concatenateStrings (const deque <string>& stringList)
{
	string result;
	for (deque <string> ::const_iterator it = stringList.begin(); it != stringList.end(); ++it)
		result += string (result == "" ? "" : ", ") + *it;
	return result;
}

static bool isCustomModifier (const string& modifierName)
{
	return modifierName != "nec"
	    && modifierName != "poss"
	    && modifierName != "d"
	    && modifierName != "f"
	    && modifierName != "g"
	    && modifierName != "h"
	    && modifierName != "hs"
	    && modifierName != "j"
	    && modifierName != "js"
	;
}

static string modifierFunction (const string& modifierName)
{
	return isCustomModifier (modifierName) ? modifierName : (IFSQL_SCHEMA_NAME".modif_" + modifierName);
}

static string modifierFunctionCall (const string& membership, const string& modifierName, const string& modifierParams)
{
	return modifierFunction (modifierName) + "(" + membership 
		+ (modifierParams == "" ? string () : (", " + modifierParams))
		+ ")";
}

static bool isRelationCrisp (const string& relname)
{
	static const char* crispPrefixes[] = { "sql_", "pg_", "crisp_" };
	
	const char* tblname = relname.c_str();
	string::size_type pos;

	if ( 
		(pos = relname.find_last_of ('.')) != string::npos
			&&
		(pos = relname.find_first_not_of (" \t\r\n", pos + 1)) != string::npos
	)
			tblname += pos;

	for (int i = 0; i < sizeof(crispPrefixes) / sizeof(*crispPrefixes); ++i)
		if (strncmp (tblname, crispPrefixes[i], strlen (crispPrefixes[i])) == 0)
			return true;
	return false;
}

void ifsql_parser_init (void)
{
	stmtVarsStack.clear ();
	pushStmtVars (stUnknown);
}

#include "scan.cpp"

